import itertools
from collections import OrderedDict

class _GridSearchMixin(object):
  """
  Mixin for grid search functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_GridSearchMixin, self).__init__()
    return

  def get_grid_iterations(self, params_grid, exceptions=None, fixed=None, priority_keys=None, verbose=2, **kwargs):
    """
    this function si a simple wrapper around basic grid search utilities

    Parameters:
    -----------
    params_grid : the actual grid dict where each param is a list

    exceptions : list of exceptions - see `grid_pos_to_params`

    priority_keys : list of keys on which the iterations will be grouped.

    verbose : int, optional
      Logging verbosity. Set 0 for no prints. Set 1 for summar prints. Set >=2 for DEBUG prints.
      Verbosity >= 2 accomodates the old parameter DEBUG.
      Default is 2.

    **kwargs:
      Available for backward compatibility.

    Returns:
    --------
      list with dictionaries
    """

    if kwargs.get('DEBUG', False):
      verbose = 2

    if verbose >= 1:
      self.P("Generating grid-search iterations from {} params dictionary".format(len(params_grid)))

    combs, params = self.grid_dict_to_values(params_grid=params_grid, priority_keys=priority_keys)
    iterations = []
    skipped = []
    n_options = len(combs)
    for i in range(n_options):
      comb = combs[i]
      func_args = self.grid_pos_to_params(comb, params, exceptions=exceptions, fixed=fixed)
      if func_args is None:
        skipped.append(self.grid_pos_to_params(comb, params))
      else:
        iterations.append(func_args)

    if verbose >= 1:
      self.P("  Generating {}/{} options.".format(len(iterations), n_options))

    if verbose >= 2:
      self.P("    Exclusions ({}):".format(len(skipped)))
      for excl in skipped:
        self.P("      {}".format(excl))
    return iterations

  def grid_dict_to_values(self, params_grid, priority_keys=None, sort_keys=True):
    """
    method to convert a grid serach dict into a list of all combinations
    returns combinations and param names for each combination entry
    """
    priority_keys = priority_keys or []
    all_keys = list(params_grid.keys())
    if sort_keys:
      priority_keys.sort()
      all_keys.sort()
    # endif sort_keys

    if len(priority_keys) > 0:
      non_priority_keys = list(set(all_keys) - set(priority_keys))
      if sort_keys:
        non_priority_keys.sort()
      # endif sort_keys
    else:
      non_priority_keys = all_keys

    assert len(set(priority_keys) & set(all_keys)) == len(priority_keys)

    params = []
    values = []
    for k in priority_keys + non_priority_keys:
      params.append(k)
      assert type(params_grid[k]) is list, 'All grid-search params must be lists. Error: {}'.format(k)
      values.append(params_grid[k])
    combs = list(itertools.product(*values))
    return combs, params

  @staticmethod
  def grid_pos_to_params(grid_data, params, exceptions=None, fixed=None):
    """
    converts a grid search combination to a dict for callbacks that expect kwargs

    exceptions: list of dicts where if certain conditions are met then we skip the position by
                returning `None` (each dict must contain a couple params and each with a list of
                excepted values - any combination will generate an exception)
    """
    func_kwargs = {}

    for j, k in enumerate(params):
      func_kwargs[k] = grid_data[j]

    if exceptions is not None:
      for exception in exceptions:
        cond = 0
        for k, v in exception.items():
          if type(v) not in [list, tuple]:
            raise ValueError("Provided exceptions must be in list format")
          if func_kwargs[k] in v:
            cond += 1
        if cond == len(exception):
          # exception met
          return None

    if fixed is not None:
      for fix in fixed:
        cond = 0
        for k,v in fix.items():
          if type(v) not in [list, tuple]:
            raise ValueError("Provided fixed params must be in list format")
          if func_kwargs[k] in v:
            cond += 1
          #endif
        #endfor
        if not (cond == 0 or cond == len(fix)):
          return None

    return func_kwargs

  @staticmethod
  def grid_dict_to_results(params_grid, value_keys):
    """
     will take a grid search dict and return a results dict
     for later dataframe conversion
     inputs:
       params_grid: grid search dict
       value_keys : the name (str) or names (list, tuple) of the test oucome
                 such as "recall" or "accuracy" or "no_episodes"
     returns:
       dict of empty lists
    """
    if type(value_keys) is str:
      value_keys = [value_keys]
    dict_results = OrderedDict()
    for vkey in value_keys:
      dict_results[vkey] = []
    for k in params_grid:
      dict_results[k] = []
    return dict_results

  @staticmethod
  def add_results(d_results, d_grid_pos, values, value_keys):
    """
     adds results to a existing results dict see `grid_dict_to_results`
     inputs:
       d_grid_pos : dict returned from grid_pos_to_params
       d_results : results dict

    """
    value_keys = [value_keys] if type(value_keys) == str else value_keys
    values = [values] if type(values) in [int, float] else values
    for k in d_grid_pos:
      d_results[k].append(d_grid_pos[k])
    for i, k in enumerate(value_keys):
      d_results[k].append(values[i])
    return d_results

  def grid_dict_to_generator(self, dict_grid, priority_keys=None):
    _combs, _params = self.grid_dict_to_values(params_grid=dict_grid, priority_keys=priority_keys)
    for _c in _combs:
      dict_pos = self.grid_pos_to_params(_c, _params)
      yield dict_pos





import abc
import base64
import json
import traceback

from naeural_core import Logger
from ratio1 import BaseDecentrAIObject
from naeural_core.local_libraries import _ConfigHandlerMixin

class FlaskWorker(BaseDecentrAIObject, _ConfigHandlerMixin):

  """
  Base class for any worker / endpoint business logic
  """

  def __init__(self, log : Logger,
               default_config,
               verbosity_level,
               worker_id,
               upstream_config=None,
               **kwargs):

    """
    Parameters:
    -----------
    log : Logger, mandatory

    default_config: dict, mandatory
      The default configuration of the worker.
      See `libraries.model_server_v2.server -> create_worker` to see the entire flow; it calls `_get_module_name_and_class`
      and searches for a `_CONFIG` in the module with the implementation and passes the value of `_CONFIG` as `default_config`

    verbosity_level: int, mandatory
      A threshold that controls the verbosity - can use it in any implementation

    worker_id : int, mandatory
      The id of the worker.

    upstream_config: dict, optional
      The upstream configuration that comes from a configuration file of the process; this `upstream_config` is merged with `default_config`
      in order to compute the final config
      The default is None ({})
    """

    self._default_config = default_config
    self._upstream_config_params = upstream_config or {}
    self.config_worker = None
    self._worker_id = worker_id
    self.__last_query = None

    self._verbosity_level = verbosity_level

    self._counter = None
    self.__encountered_error = None
    prefix_log = kwargs.pop('prefix_log', '[FSKWKR]')
    super(FlaskWorker, self).__init__(log=log, maxlen_notifications=1000, prefix_log=prefix_log, **kwargs)
    return

  def startup(self):
    super().startup()
    self.config_worker = self._merge_prepare_config()
    self.setup_config_and_validate(self.config_worker)
    self._load_model()
    return

  @abc.abstractmethod
  def _load_model(self):
    """
    Implement this method in sub-class - custom logic for loading the model. If the worker has no model, then implement
    it as a simple `return`.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _pre_process(self, inputs):
    """
    Implement this method in sub-class - custom logic for pre-processing the inputs that come from the user

    Parameters:
    -----------
    inputs: dict, mandatory
      The request json

    Returns:
    -------
    prep_inputs:
      Any object that will be used in `_predict` implementation
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _predict(self, prep_inputs):
    """
    Implement this method in sub-class - custom logic for predict

    Parameters:
    -----------
    prep_inputs:
      Any object returned by `_pre_process`

    preds:
      Any object that will be used in `_post_process` implementation
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _post_process(self, pred):
    """
    Implement this method in sub-class - custom logic for post processing the predictions
    (packing the output that goes to the end-user)

    Parameters:
    ----------
    pred:
      Any object returned by `_predict`

    answer: dict
      The answer that goes to the end-user
    """
    raise NotImplementedError

  @staticmethod
  def __err_dict(err_type, err_file, err_func, err_line, err_msg):
    return {
      'ERR_TYPE' : err_type,
      'ERR_MSG' : err_msg,
      'ERR_FILE' : err_file,
      'ERR_FUNC' : err_func,
      'ERR_LINE' : err_line
    }


  def __pre_process(self, inputs):
    try:
      prep_inputs = self._pre_process(inputs)
    except:
      err_dict = self.__err_dict(*self.log.get_error_info(return_err_val=True))
      msg = 'Exception in _pre_process:\n{}'.format(err_dict)
      self.__encountered_error = err_dict #['ERR_MSG']
      
      self._create_notification(
        notif='exception',
        msg=msg,
      )
      return

    return prep_inputs

  def __predict(self, prep_inputs):
    if prep_inputs is None:
      return

    try:
      pred = self._predict(prep_inputs)
    except:
      exc_info = traceback.format_exc()
      err_dict = self.__err_dict(*self.log.get_error_info(return_err_val=True))
      self.__encountered_error = err_dict #['ERR_MSG']
      msg = 'Exception in _predict:\n{}'.format(err_dict)
      self._create_notification(
        notif='exception',
        msg=msg
      )
      self.P("{}:\n{}".format(msg, exc_info), color='r')
      pred = None
    return pred

  def __post_process(self, pred):
    if pred is None:
      return

    try:
      answer = self._post_process(pred)
    except:
      err_dict = self.__err_dict(*self.log.get_error_info(return_err_val=True))
      self.__encountered_error = err_dict #['ERR_MSG']
      msg = 'Exception in _post_process\n{}'.format(err_dict)
      self._create_notification(
        notif='exception',
        msg=msg,
      )
      return

    return answer
  
  
  def get_last_query(self):
    return self.__last_query
  

  def execute(self, inputs, counter):
    """
    The method exposed for execution.

    Parameters:
    ----------
    inputs: dict, mandatory
      The request json

    counter: int, mandatory
      The call id

    Returns:
    --------
    answer: dict
      The answer that goes to the end-user
    """
    self._counter = counter
    self.__encountered_error = None

    base64_keys = inputs.pop('BASE64_KEYS', [])
    base64_outputs = inputs.pop('BASE64_OUTPUTS', [])
    encoding = inputs.pop('ENCODING', 'ansi')
    for k in base64_keys:
      if k in inputs:
        inputs[k] = base64.b64decode(inputs[k]).decode(encoding)
      else:
        self.P("Key {} sent in 'BASE64_KEYS' does not exist in input", color='e')
    #endfor

    self.__last_query = inputs
    
    prep_inputs = self.__pre_process(inputs)
        
    pred = self.__predict(prep_inputs)

    answer = self.__post_process(pred)

    for k in base64_outputs:
      if k in answer:
        if isinstance(answer[k], str):
          answer[k] = base64.b64encode(answer[k].encode(encoding)).decode()
        else:
          answer[k] = base64.b64encode(json.dumps(answer[k]).encode(encoding)).decode()
      else:
        self.P("Key {} sent in 'BASE64_OUTPUTS' does not exist in answer", color='e')
    #endfor

    if self.__encountered_error:
      answer = {'{}_ERROR'.format(self.__class__.__name__) : self.__encountered_error}
      self.P("Worker {}:{} execute error: {}".format(
        self.__class__.__name__, self._worker_id, self.__encountered_error), color='r'
      )

    return answer

  def _create_notification(self, notif, msg, info=None, stream_name=None, **kwargs):
    msg = (self._counter or "INIT", msg)
    super()._create_notification(notif=notif, msg=msg, info=info, stream_name=stream_name, **kwargs)
    return



from naeural_core.business.base import BasePluginExecutor as BasePlugin

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  'VALIDATION_RULES' : {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },

  'AI_ENGINE' : 'th_training',
  'STARTUP_AI_ENGINE_PARAMS' : {
    'PIPELINE_SIGNATURE' : None,
    'PIPELINE_CONFIG' : {},
  },

  "DESCRIPTION": "",
  "OBJECTIVE_NAME": "",
  'PLUGIN_LOOP_RESOLUTION' : 1/60,
  'ALLOW_EMPTY_INPUTS' : False,
  'AUTO_DEPLOY' : {},
}


class GeneralTrainingProcessPlugin(BasePlugin):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self.training_output = None
    self.performed_final = False

    self.weights_uri = {'URL': '', 'CLOUD_PATH' : ''}
    self.trace_uri = {'URL': '', 'CLOUD_PATH' : ''}
    self.training_output_uri = {'URL': '', 'CLOUD_PATH': ''}
    self.inference_config_uri = {'URL': '', 'CLOUD_PATH': ''}
    super(GeneralTrainingProcessPlugin, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    assert self.cfg_startup_ai_engine_params.get('PIPELINE_SIGNATURE', None) is not None
    assert bool(self.cfg_startup_ai_engine_params.get('PIPELINE_CONFIG', {}))
    if self.cfg_auto_deploy:
      assert self.inspect.getsource(self.prepare_inference_plugin_config) != self.inspect.getsource(GeneralTrainingProcessPlugin.prepare_inference_plugin_config),\
             "When auto deploy is configured, `_prepare_inference_plugin_config` should be defined"

      assert self.inspect.getsource(self.auto_deploy) != self.inspect.getsource(GeneralTrainingProcessPlugin.auto_deploy),\
             "When auto deploy is configured, `_auto_deploy` should be defined"
    #endif

    return

  @property
  def cfg_startup_ai_engine_params(self):
    return self._instance_config['STARTUP_AI_ENGINE_PARAMS']

  @property
  def cfg_auto_deploy(self):
    return self._instance_config.get('AUTO_DEPLOY', {})

  def cloud_path_to_url(self, cloud_path):
    return f'minio:{cloud_path}'

  def get_instance_config(self):
    """
    This method is overriden to allow for multiple training jobs on the same node
    Returns
    -------
    res - dict, the instance config with the MODEL_INSTANCE_ID added
    """
    current_instance_config = super().get_instance_config()
    curr_ai_engine = current_instance_config.get('AI_ENGINE', 'th_training')
    curr_startup_params = current_instance_config.get('STARTUP_AI_ENGINE_PARAMS', {})
    model_instance_id = curr_startup_params.get('MODEL_INSTANCE_ID', None)
    if model_instance_id is None:
      if '?' in curr_ai_engine:
        curr_ai_engine, model_instance_id = curr_ai_engine.split('?')
        current_instance_config['AI_ENGINE'] = curr_ai_engine
      else:
        model_instance_id = current_instance_config.get('INSTANCE_ID', 'default')
      # endif model_instance_id provided in AI_ENGINE
      current_instance_config['STARTUP_AI_ENGINE_PARAMS'] = {
        **curr_startup_params,
        'MODEL_INSTANCE_ID': model_instance_id
      }
    # endif model_instance_id not provided

    return current_instance_config

  def prepare_inference_plugin_config(self) -> dict:
    return {}

  def auto_deploy(self):
    return

  def on_training_finish(self, model_id):
    training_subdir = 'training'
    # Model
    best_weights = (self.training_output['STATUS']['BEST'] or {}).get('best_file')
    traced_model_path = (self.training_output['METADATA'] or {}).get('MODEL_EXPORT')
    if best_weights is None:
      self.P("Best weights not found, aborting upload", color='r')
      return False
    files = {'weights': best_weights, 'trace': traced_model_path}
    if traced_model_path is None:
      self.P("Traced model path not found, uploading only the weights of best model", color='y')
      files = {'weights': best_weights}
    # endif trace not found
    uris = {}
    for file_id, file_path in files.items():
      uris[file_id] = {'CLOUD_PATH': f'TRAINING/{model_id}/{file_id}/{file_path.split(self.os_path.sep)[-1]}'}
      url, _ = self.upload_file(
        file_path=file_path,
        target_path=uris[file_id]['CLOUD_PATH'],
        force_upload=True,
      )
      uris[file_id]['URL'] = url
    # endfor files
    self.weights_uri = uris['weights']
    self.trace_uri = uris.get('trace',  {'URL': None, 'CLOUD_PATH': None})
    # Training output
    json_name = f'{model_id}_training_output.json'
    path_training_output = self.os_path.join(training_subdir, json_name)
    # First we save it, so we can upload it after
    json_path = self.diskapi_save_json_to_output(dct=self.training_output, filename=path_training_output)
    self.training_output_uri['CLOUD_PATH'] = 'TRAINING/{}/{}'.format(model_id, json_name)
    url, _ = self.upload_file(
      file_path=json_path,
      target_path=self.training_output_uri['CLOUD_PATH'],
      force_upload=True
    )
    self.training_output_uri['URL'] = url

    if bool(self.cfg_auto_deploy):
      # Inference config
      dct_inference_config = self.prepare_inference_plugin_config()
      json_name = f'{model_id}_inference_config.json'
      path_inference_config = self.os_path.join(training_subdir, json_name)
      json_path = self.diskapi_save_json_to_output(dct=dct_inference_config, filename=path_inference_config)
      self.inference_config_uri['CLOUD_PATH'] = 'TRAINING/{}/{}'.format(model_id, json_name)
      url, _ = self.upload_file(
        file_path=json_path,
        target_path=self.inference_config_uri['CLOUD_PATH'],
        force_upload=True
      )
      self.inference_config_uri['URL'] = url
    #endif

    return True

  def _process(self):
    self.training_output = self.dataapi_specific_struct_data_inferences(idx=0, how='list', raise_if_error=True)
    assert len(self.training_output) == 1
    self.training_output = self.training_output[0]
    assert isinstance(self.training_output, dict)
    assert 'STATUS' in self.training_output
    has_finished = self.training_output.get('HAS_FINISHED', False)
    payload_kwargs = {
      'is_status': True,
      'train_status': self.training_output,
      'description': self.cfg_description,
      'objective_name': self.cfg_objective_name,
    }
    if self.training_output['STATUS'] != 'WAITING':
      payload_kwargs['job_status'] = 'Training' if not has_finished else 'Trained'
    save_payload_json = False
    model_id = None
    if has_finished:
      success = self.performed_final
      if not self.performed_final:
        self.P("Training has finished", color='g')
        model_id = '{}_{}'.format(self.log.session_id, self.training_output['METADATA']['MODEL_NAME'])
        save_payload_json = True
        success = self.on_training_finish(model_id=model_id)

        if success:
          if bool(self.cfg_auto_deploy):
            self.auto_deploy()

          self.performed_final = True
        else:
          self.P(f"Training finished but upload failed. Will re-attempt...", color='r')
      # endif performed final
      if success:
        train_final_kwargs = {
          'MODEL_URI': self.weights_uri,
          'TRACE_URI': self.trace_uri,
          'TRAINING_OUTPUT_URI': self.training_output_uri,
          'INFERENCE_CONFIG_URI': self.inference_config_uri,
        }
        payload_kwargs['TRAIN_FINAL'] = train_final_kwargs
      # endif success
    # endif training finished

    payload = self._create_payload(**payload_kwargs)
    if save_payload_json:
      self.diskapi_save_json_to_data(
        dct=payload.to_dict(),
        filename=f'training/{model_id}_golden_payload.json'
      )
    # endif payload needs to be saved

    return payload




from naeural_core.heavy_ops.base import BaseHeavyOp
from naeural_core import constants as ct
from naeural_core.utils.emailer import send_email
import json

CONFIG = {
  'IDLE_THREAD_SLEEP_TIME' : 2,
}

class SendMailHeavyOp(BaseHeavyOp):
  """
  This op handles payloads that have _H_SEND_EMAIL set to True - and will send email
  as a threaded heavy op
  """

  def __init__(self, **kwargs):
    super(SendMailHeavyOp, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    assert self.comm_async

  def _register_payload_operation(self, payload):
    dct = payload.copy()
    payload.pop('_H_SEND_EMAIL', False)
    payload.pop('_H_EMAIL_CONFIG', None)
    payload.pop('_H_EMAIL_MESSAGE', None)
    payload.pop('_H_EMAIL_SUBJECT', None)
    return dct

  def _process_dct_operation(self, dct):
    bool_send_email = dct.pop('_H_SEND_EMAIL', False)
    email_config = dct.pop('_H_EMAIL_CONFIG', None)
    subject = dct.get('_H_EMAIL_SUBJECT', None)
    message = dct.get('_H_EMAIL_MESSAGE', None)

    if bool_send_email and email_config is not None and subject is not None:
      user = email_config.get(ct.EMAIL_NOTIFICATION.USER, ct.EMAIL_NOTIFICATION.DEFAULT_USER)
      pwd = email_config.get(ct.EMAIL_NOTIFICATION.PASSWORD, ct.EMAIL_NOTIFICATION.DEFAULT_PASSWORD)
      server = email_config.get(ct.EMAIL_NOTIFICATION.SERVER, ct.EMAIL_NOTIFICATION.DEFAULT_SERVER)
      port = email_config.get(ct.EMAIL_NOTIFICATION.PORT, ct.EMAIL_NOTIFICATION.DEFAULT_PORT)
      dest = email_config.get(ct.EMAIL_NOTIFICATION.DESTINATION)

      dct.pop('IMG', None)

      if message is None:
        message = json.dumps(dct, indent=4)

      if dest is not None:
        res, logs = send_email(
          log=self.log,
          server=server,
          port=port,
          user=user,
          password=pwd,
          destination=dest,
          subject=subject,
          msg=message,
          return_logs=True
        )

        for dl in logs:
          self._create_notification(
            notif='LOG',
            msg=dl['msg']
          )
        #endfor
      #endif

    return



VER = '0.0.1'

from tensorflow.python.framework import ops
from tensorflow.python.ops import state_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.framework import constant_op
import tensorflow as tf


class COCOB(tf.keras.optimizers.Optimizer):
  def __init__(self, alpha=100, use_locking=False, name='COCOB', **kwargs):
    '''
    constructs a new COCOB optimizer - copyright 2017 Francesco Orabona
    https://github.com/bremen79/cocob
    '''
    super(COCOB, self).__init__(name, **kwargs)
    self._alpha = alpha
    self._use_locking = use_locking

  def _create_slots(self, var_list):
    for v in var_list:
      with ops.colocate_with(v):
        gradients_sum = constant_op.constant(0,
                                             shape=v.get_shape(),
                                             dtype=v.dtype.base_dtype)
        grad_norm_sum = constant_op.constant(0,
                                             shape=v.get_shape(),
                                             dtype=v.dtype.base_dtype)
        L = constant_op.constant(1e-8, shape=v.get_shape(), dtype=v.dtype.base_dtype)
        tilde_w = constant_op.constant(0.0, shape=v.get_shape(), dtype=v.dtype.base_dtype)
        reward = constant_op.constant(0.0, shape=v.get_shape(), dtype=v.dtype.base_dtype)

      self.add_slot(v, "L", initializer=L)
      self.add_slot(v, "grad_norm_sum", initializer=grad_norm_sum)
      self.add_slot(v, "gradients_sum", initializer=gradients_sum)
      self.add_slot(v, "tilde_w", initializer=tilde_w)
      self.add_slot(v, "reward", initializer=reward)

  def _apply_dense(self, grad, var):
    gradients_sum = self.get_slot(var, "gradients_sum")
    grad_norm_sum = self.get_slot(var, "grad_norm_sum")
    tilde_w = self.get_slot(var, "tilde_w")
    L = self.get_slot(var, "L")
    reward = self.get_slot(var, "reward")

    L_update = tf.maximum(L, tf.abs(grad))
    gradients_sum_update = gradients_sum + grad
    grad_norm_sum_update = grad_norm_sum + tf.abs(grad)
    reward_update = tf.maximum(reward - grad * tilde_w, 0)
    new_w = -gradients_sum_update / (
              L_update * (tf.maximum(grad_norm_sum_update + L_update, self._alpha * L_update))) * (
                      reward_update + L_update)
    var_update = var - tilde_w + new_w
    tilde_w_update = new_w

    gradients_sum_update_op = state_ops.assign(gradients_sum, gradients_sum_update)
    grad_norm_sum_update_op = state_ops.assign(grad_norm_sum, grad_norm_sum_update)
    var_update_op = state_ops.assign(var, var_update)
    tilde_w_update_op = state_ops.assign(tilde_w, tilde_w_update)
    L_update_op = state_ops.assign(L, L_update)
    reward_update_op = state_ops.assign(reward, reward_update)

    return control_flow_ops.group(*[gradients_sum_update_op,
                                    var_update_op,
                                    grad_norm_sum_update_op,
                                    tilde_w_update_op,
                                    reward_update_op,
                                    L_update_op])

  def _apply_sparse(self, grad, var):
    return self._apply_dense(grad, var)

  def _resource_apply_dense(self, grad, handle):
    return self._apply_dense(grad, handle)

  def _resource_apply_sparse(self, grad, handle, indices, apply_state=None):
    raise NotImplementedError()

  def get_config(self):
    config = {
      'alpha': self._alpha,
      'use_locking': self._use_locking
    }
    base_config = super().get_config()
    cfg = dict(list(base_config.items()) + list(config.items()))
    return cfg

  @classmethod
  def from_config(cls, config, custom_objects=None):
    print("\x1b[1;33mLoading COCOB optimizer\x1b[0m")
    return cls(**config)


from naeural_core.data.base import AbstractMapReduceDataCapture

_CONFIG = {
  **AbstractMapReduceDataCapture.CONFIG,
  
  'VALIDATION_RULES' : {
    **AbstractMapReduceDataCapture.CONFIG['VALIDATION_RULES'],
  },
}

class BasicMapReduceDataCapture(AbstractMapReduceDataCapture):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    super(BasicMapReduceDataCapture, self).__init__(**kwargs)
    return

  def _init_map(self):
    return

  def _release_map(self):
    return

  def _maybe_reconnect_map(self):
    return

  def _run_data_aquisition_step_map(self):
    workers = self._workers[:len(self.cfg_url)]
    self._add_inputs(
      [
        self._new_input(struct_data=dict(url_map=self.cfg_url, workers=workers))
      ]
    )

    self._finished_map = True
    return



import tensorflow as tf

def luong_attention(q, k, v, ffn=None):
  """
  Calculate the attention weights.
  q, k, v must have matching leading dimensions.
  k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.

  Args:
    q: query shape == (..., seq_len_q, depth_q)
    k: key shape == (..., seq_len_k, depth_k)
    v: value shape == (..., seq_len_v, depth_v)
    ffn: a feed-forward network object that transforms the query (q) in order to be
         dot-multiplied with the key (k). The last layer of the `ffn` should have
         `units=depth_k`.
         The default value is None. In this case, `ffn` is a Dense layer with `units=depth_k`

    Returns:
      attn, attn_sm
    """

  depth_k = k.shape[-1]
  if ffn is None:
    layers = [tf.keras.layers.Dense(units=depth_k,
                                    name='q_transform_{}'.format(depth_k))]
    ffn = tf.keras.models.Sequential(layers)
  #endif

  assert ffn.layers[-1].units == depth_k

  q_transformed = ffn(q) # (batch_size, depth_k)
  q_exp = tf.expand_dims(q_transformed, axis=1) # (batch_size, 1, depth_k)
  query_keys = tf.keras.backend.batch_dot(k, q_exp, axes=2) # (batch_size, seq_len_k, 1)

  attn_logits = query_keys
  attn_sm = tf.keras.activations.softmax(attn_logits, axis=1)
  attn = tf.reduce_sum(attn_sm * v, axis=1, keepdims=False) # (batch_size, depth_v)

  return attn, attn_sm

def bahdanau_one_attention(q, k, v, ffn=None):
  """
  Calculate the attention weights.
  q, k, v must have matching leading dimensions.
  k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.

  Args:
    q: query shape == (..., seq_len_q, depth_q)
    k: key shape == (..., seq_len_k, depth_k)
    v: value shape == (..., seq_len_v, depth_v)
    ffn: a feed-forward network object that transforms the result of query (q)
         applied on keys (k). The last layer of the `ffn` should have `units=1`.
         The default value is None. In this case, `ffn` is a Dense layer with `units=1`

    Returns:
      attn, attn_sm
    """
  if ffn is None:
    layers = [tf.keras.layers.Dense(units=1,
                                    name='qk_transform_{}'.format(1))]
    ffn = tf.keras.models.Sequential(layers)
  #endif

  assert ffn.layers[-1].units == 1

  seq_len_k = k.shape[-2]

  q_exp = tf.expand_dims(q, axis=1) # (batch_size, 1, depth_q)
  repeated_query = q_exp + tf.zeros((seq_len_k, 1)) # (batch_size, seq_len_k, depth_q)
  query_keys = tf.concat([repeated_query, k], axis=-1) # (batch_size, seq_len_k, depth_q + depth_k)

  attn_logits = ffn(query_keys) # (batch_size, seq_len_k, 1)
  attn_sm = tf.keras.activations.softmax(attn_logits, axis=1)
  attn = tf.reduce_sum(attn_sm * v, axis=1, keepdims=False)

  return attn, attn_sm

def bahdanau_two_attention(q, k, v, ffn=None, W1=None, W2=None):
  """
  Calculate the attention weights.
  q, k, v must have matching leading dimensions.
  k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.

  Args:
    q: query shape == (..., seq_len_q, depth_q)
    k: key shape == (..., seq_len_k, depth_k)
    v: value shape == (..., seq_len_v, depth_v)
    ffn: a feed-forward network object that transforms the result of query (q)
         applied on keys (k). The last layer of the `ffn` should have `units=1`.
         The default value is None. In this case, `ffn` is a Dense layer with `units=1`
    W1: a Dense object that transforms the key in order to be summed.
        The default value is None. In this case, `W1` is a Dense layer with `units=depth_k`.
    W2: a Dense object that transforms the query in order to be summed.
        The default value is None. In this case, `W1` is a Dense layer with `units=depth_k`.

    Returns:
      attn, attn_sm
    """
  if ffn is None:
    layers = [tf.keras.layers.Dense(units=1,
                                    name='qk_transform_{}'.format(1))]
    ffn = tf.keras.models.Sequential(layers)
  #endif

  assert ffn.layers[-1].units == 1

  depth_k = k.shape[-1]
  if W1 is None:
    W1 = tf.keras.layers.Dense(units=depth_k, use_bias=False,
                               name='k_transform_{}'.format(depth_k))
  if W2 is None:
    W2 = tf.keras.layers.Dense(units=depth_k, use_bias=False,
                               name='q_transform_{}'.format(depth_k))
  query_keys = tf.nn.tanh(W1(k) + W2(q)) # (batch_size, seq_len_k, depth_k)
  attn_logits = ffn(query_keys) # (batch_size, seq_len_k, 1)
  attn_sm = tf.keras.activations.softmax(attn_logits, axis=1)
  attn = tf.reduce_sum(attn_sm * v, axis=1, keepdims=False)

  return attn, attn_sm


class LuongAttention(tf.keras.layers.Layer):
  def __init__(self, key_depth, **kwargs):
    super(LuongAttention, self).__init__(**kwargs)
    self.key_depth = key_depth

    layers = [tf.keras.layers.Dense(units=self.key_depth,
                                    name='q_transform_{}'.format(self.key_depth))]
    self.ffn = tf.keras.models.Sequential(layers)
    return

  def call(self, inputs):
    q,k,v = inputs
    return luong_attention(q=q, k=k, v=v, ffn=self.ffn)

  def get_config(self):
    config = {
      'key_depth' : self.key_depth,
    }
    base_config = super(LuongAttention, self).get_config()
    cfg =  dict(list(base_config.items()) + list(config.items()))
    return cfg


class BahdanauAttention(tf.keras.layers.Layer):
  def __init__(self, key_depth, mode=1, **kwargs):
    assert mode in [1,2]

    super(BahdanauAttention, self).__init__(**kwargs)
    self.key_depth = key_depth
    self.mode = mode

    layers = [tf.keras.layers.Dense(units=1,
                                    name='qk_transform_{}'.format(1))]
    self.ffn = tf.keras.models.Sequential(layers)

    self.W1 = tf.keras.layers.Dense(units=self.key_depth, use_bias=False,
                                    name='k_transform_{}'.format(self.key_depth))
    self.W2 = tf.keras.layers.Dense(units=self.key_depth, use_bias=False,
                                    name='q_transform_{}'.format(self.key_depth))
    return

  def call(self, inputs):
    q,k,v = inputs
    if self.mode == 1:
      return bahdanau_one_attention(q=q, k=k, v=v, ffn=self.ffn, W1=self.W1, W2=self.W2)
    elif self.mode == 2:
      return bahdanau_two_attention(q=q, k=k, v=v, ffn=self.ffn, W1=self.W1, W2=self.W2)

  def get_config(self):
    config = {
      'key_depth': self.key_depth,
      'mode' : self.mode,
    }
    base_config = super(BahdanauAttention, self).get_config()
    cfg =  dict(list(base_config.items()) + list(config.items()))
    return cfg


def scaled_dot_product_attention(q, k, v, mask=None):
  """
  Calculate the attention weights.
  q, k, v must have matching leading dimensions.
  k, v must have matching penultimate dimension, i.e.: seq_len_k = seq_len_v.
  The mask has different shapes depending on its type (padding or look ahead)
  but it must be broadcastable for addition.

  Args:
    q: query shape == (..., seq_len_q, depth)
    k: key shape == (..., seq_len_k, depth)
    v: value shape == (..., seq_len_v, depth_v)
    mask: Float tensor with shape broadcastable
          to (..., seq_len_q, seq_len_k). Defaults to None.

  Returns:
    output, attention_weights
  """

  matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)

  # The dot-product attention is scaled by a factor of square root of the depth.
  # This is done because for large values of depth, the dot product grows large
  # in magnitude pushing the softmax function where it has small gradients
  # resulting in a very hard softmax.
  dk = tf.cast(tf.shape(k)[-1], tf.float32)
  scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

  # add the mask to the scaled tensor.
  if mask is not None:
    scaled_attention_logits += (mask * -1e9)

  # softmax is normalized on the last axis (seq_len_k)
  attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)  # (..., seq_len_q, seq_len_k)

  output = tf.matmul(attention_weights, v)  # (..., seq_len_q, depth_v)

  return output, attention_weights


class MultiHeadAttention(tf.keras.layers.Layer):
  def __init__(self, d_model, num_heads):
    super(MultiHeadAttention, self).__init__()
    self.num_heads = num_heads
    self.d_model = d_model

    assert d_model % self.num_heads == 0

    self.depth = d_model // self.num_heads

    self.wq = tf.keras.layers.Dense(d_model)
    self.wk = tf.keras.layers.Dense(d_model)
    self.wv = tf.keras.layers.Dense(d_model)

    self.dense = tf.keras.layers.Dense(d_model)
    return

  def split_heads(self, x, batch_size):
    """Split the last dimension into (num_heads, depth).
    Transpose the result such that the shape is (batch_size, num_heads, seq_len, depth)
    """
    x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
    return tf.transpose(x, perm=[0, 2, 1, 3])

  def call(self, q, k, v, mask=None):
    batch_size = tf.shape(q)[0]

    # Instead of one single attention head, Q, K, and V are split into multiple heads
    # because it allows the model to jointly attend to information at different
    # positions from different representational spaces. After the split each
    # head has a reduced dimensionality, so the total computation cost is the same
    # as a single head attention with full dimensionality.
    q = self.wq(q)  # (batch_size, seq_len, d_model)
    k = self.wk(k)  # (batch_size, seq_len, d_model)
    v = self.wv(v)  # (batch_size, seq_len, d_model)

    q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
    k = self.split_heads(k, batch_size)  # (batch_size, num_heads, seq_len_k, depth)
    v = self.split_heads(v, batch_size)  # (batch_size, num_heads, seq_len_v, depth)

    # scaled_attention.shape == (batch_size, num_heads, seq_len_q, depth)
    # attention_weights.shape == (batch_size, num_heads, seq_len_q, seq_len_k)
    scaled_attention, attention_weights = scaled_dot_product_attention(
      q, k, v, mask)

    scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])  # (batch_size, seq_len_q, num_heads, depth)

    concat_attention = tf.reshape(scaled_attention,
                                  (batch_size, -1, self.d_model))  # (batch_size, seq_len_q, d_model)

    output = self.dense(concat_attention)  # (batch_size, seq_len_q, d_model)

    return output, attention_weights


if __name__ == '__main__':
  import numpy as np
  query = tf.constant(np.array([[0.23, 0.25, 1.43, 1.45, 0.01, 0.45, 0.73, 0.88],
                                [0.15, 0.77, 1.41, 0.03, 0.12, 0.17, 1.17, 1.12]]))
  enc_states = np.array([[[1, 2, 3, 4, 5], [0, 0, 0, 0, 0], [1, 2, 3, 4, 5], [0, 0, 0, 0, 0]],
                         [[1, 2, 3, 4, 5], [0, 0, 0, 0, 0], [1, 2, 3, 4, 5], [0, 0, 0, 0, 0]]])
  key = tf.constant(enc_states, dtype=tf.float32)
  value = tf.constant(enc_states, dtype=tf.float32)

  a, b = luong_attention(q=query, k=key, v=value)



"""
Example pipeline configuration:

{
    "CAP_RESOLUTION": 20,
    "NAME": "on-demand-stream",
    "PLUGINS": [
        {
            "INSTANCES": [
                {
                    "AI_ENGINE"   : "a_sum_model",
                    "INSTANCE_ID": "default"
                }
            ],
            "SIGNATURE": "a_simple_plugin"
        }
    ],
    "TYPE": "OnDemandInput",
    "STREAM_CONFIG_METADATA" : {
      "SERVER_SIDE_PARAMS_HERE" : "1234"
    },
    "URL": 0
}

then:

{ 
  "ACTION" : "PIPELINE_COMMAND",  
  "PAYLOAD" : {
    "NAME": "on-demand-stream",
    "PIPELINE_COMMAND" : {
      "STRUCT_DATA" : [1,2,3,4]
    }
  }
}
"""

from naeural_core.data.base import DataCaptureThread

__VER__ = '0.1.0'

_CONFIG = {
  **DataCaptureThread.CONFIG,
  
  'CAP_RESOLUTION'    : 10, # default run at 10Hz
  
  'LIVE_FEED'         : False, # it should run buffered
  
  'VALIDATION_RULES'  : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}

class OnDemandInputDataCapture(
  DataCaptureThread,

  ):

  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(OnDemandInputDataCapture, self).__init__(**kwargs)
    return

  
  def startup(self):
    # NON-threaded code in startup
    super().startup()
    self._metadata.update(
      pipeline_version=__VER__,
    )
    return 
  
  
  def _init(self):
    self._iter = 0
    # not much here
    self._maybe_reconnect()    
    return
  
  @property
  def is_idle_alert(self):
    # we are never idle :)
    return False
  
  
  def _maybe_reconnect(self):
    # not much here
    if self.has_connection:
      return    
    self.has_connection = True
    return  


  def _on_pipeline_command(self, cmd_data, payload_context=None, **kwargs):
    _obs = cmd_data.get('STRUCT_DATA', None)
    _base64_img = cmd_data.get('IMG', None)
    _train = None
    if _base64_img is not None:
      _img = self.base64_to_img(_base64_img)
      self._add_inputs(
        [
          self._new_input(img=_img, struct_data=None, metadata=self._metadata.__dict__.copy(), init_data=None),
        ]
      )
      # This is done in order to not have the image received saved in the plugin config
      if 'LAST_PIPELINE_COMMAND' in self.config_data:
        obj_size = self.log.get_obj_size(self.config_data['LAST_PIPELINE_COMMAND'])
        self.P(f"Last pipeline command had size {obj_size} bytes")
        self.config_data['LAST_PIPELINE_COMMAND'] = None
        self.archive_config_keys(keys=['PIPELINE_COMMAND'], defaults=[{}])
      # endif reset last pipeline command
    else:
      self._add_inputs(
        [
          self._new_input(img=None, struct_data=_obs, metadata=self._metadata.__dict__.copy(), init_data=_train),
        ]
      )
    return  
    
    
  def _run_data_aquisition_step(self):
    self._iter += 1
    # not much here either as stuff happens in _on_pipeline_command
    return 
    
    
  def _release(self):
    return  
  
  


import torch as th

from naeural_core.serving.default_inference.th_yf8l import ThYf8l as BaseServingProcess

_CONFIG = {
  **BaseServingProcess.CONFIG,

  "DEBUG_TIMERS": False,
  'MAX_BATCH_SECOND_STAGE': 20,

  'COVERED_SERVERS': ['th_yf8l'],

  #####################
  "MSPN_URL": "minio:MSPN/ThYoloMSPN_bs1.ths",
  "MSPN_FILENAME": "ThYoloMSPN_bs1.ths",

  "MSPN_ONNX_URL" : "minio:MSPN/20240430_ThYoloMSPN_bs1.onnx",
  "MSPN_ONNX_FILENAME" : "20240430_ThYoloMSPN_bs1.onnx",
  "MSPN_BACKEND" : None,

  "MSPN_IMG_SHAPE": (256, 192),
  "NORM_MEANS": [0.406, 0.456, 0.485],
  "NORM_STDS": [0.225, 0.224, 0.229],
  "CROP_ORIGINAL": True,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

DEBUG_COVERED_SERVERS = False


class ThYfMspn(BaseServingProcess):
  CONFIG = _CONFIG

  def get_mspn_model_config(self):
    """
    The filename of the selected mspn model.
    """
    # First try to get the TensorRT/ONXN config
    if self.cfg_mspn_onnx_filename is not None:
      onnx_config = self.graph_config.get(self.cfg_mspn_onnx_filename)
      if onnx_config is not None:
        return onnx_config
    #endif onnx config
    # No TensorRT/ONNX model was selected, we must have used ths.
    return self.graph_config[self.cfg_mspn_filename]

  def __init__(self, **kwargs):
    super(ThYfMspn, self).__init__(**kwargs)
    self._has_second_stage_classifier = True
    return

  def model_call_mspn(self, th_inputs, model):
    movenet_preds = model(th_inputs)
    return self._post_process_mspn(movenet_preds, warmup=True)

  def _mspn_warmup(self):
    self.model_warmup_helper(
      model=self.mspn_model,
      model_call_method=self.model_call_mspn,
      input_shape=(3, *self.cfg_mspn_img_shape),
      max_batch_size=self.cfg_max_batch_second_stage,
      model_name='mspn'
    )
    return

  def _load_mspn_model(self):
    # Note that we use the same onnx file for trt/onnx/openvino.
    config = {
      'ths' : (self.cfg_mspn_filename, self.cfg_mspn_url),
      'trt' : (self.cfg_mspn_onnx_filename, self.cfg_mspn_onnx_url),
      'onnx' : (self.cfg_mspn_onnx_filename, self.cfg_mspn_onnx_url),
      'openvino' : (self.cfg_mspn_onnx_filename, self.cfg_mspn_onnx_url),
    }
    self.mspn_model, model_loaded_config, fn = self.prepare_model(
      backend_model_map=config,
      forced_backend=self.cfg_mspn_backend,
      return_config=True,
      batch_size=self.cfg_max_batch_second_stage
    )
    self.graph_config[fn] = model_loaded_config

    self.transforms = self.tv.transforms.Compose([
      self.tv.transforms.Resize(self.cfg_mspn_img_shape),
      self.tv.transforms.Normalize(
        mean=self.cfg_norm_means,  # cfg.INPUT.MEANS,
        std=self.cfg_norm_stds  # cfg.INPUT.STDS
      )
    ])

    self._mspn_warmup()
    return

  def _second_stage_model_load(self):
    self._load_mspn_model()
    return

  def _pre_process_images(self, images, **kwargs):
    return super(ThYfMspn, self)._pre_process_images(
      images=images,
      return_original=True,
      normalize_original=True,
      **kwargs
    )

  def _post_process(self, preds):
    yolo_preds, second_preds = preds
    lst_yolo_results = super(ThYfMspn, self)._post_process(preds)
    if second_preds is not None:
      for i, yolo_result in enumerate(lst_yolo_results):
        for j, crop_results in enumerate(yolo_result):
          if crop_results['TYPE'] == 'person':
            crop_results["JOINTS"] = second_preds[i][j]
          # endif
        # endfor
      # endfor
    # endif
    return lst_yolo_results

  @staticmethod
  def _aggregate_second_stage_batch_predict(lst_results):
    return th.cat(lst_results)

  @staticmethod
  def unravel_index(
      indices: th.LongTensor,
      shape,
  ) -> th.LongTensor:
    r"""Converts flat indices into unraveled coordinates in a target shape.

    This is a `torch` implementation of `numpy.unravel_index`.

    Args:
        indices: A tensor of indices, (*, N).
        shape: The targeted shape, (D,).

    Returns:
        unravel coordinates, (*, N, D).
    """

    shape = th.tensor(shape)
    indices = indices % shape.prod()  # prevent out-of-bounds indices

    coord = th.zeros(indices.size() + shape.size(), dtype=int)

    for i, dim in enumerate(reversed(shape)):
      coord[..., i] = indices % dim
      indices = th.div(indices, dim, rounding_mode='floor')

    return coord.flip(-1)

  def _post_process_mspn(self, outputs, kernel=5, shifts=[0.25], warmup=False):
    if not warmup:
      self._start_timer("pproc_img")

    nr_img = outputs.shape[0]
    score_maps = self.th.clone(outputs)
    score_maps = score_maps / 255 + 0.5

    kps = self.th.zeros((nr_img, self.get_mspn_model_config()['DATASET_KEYPOINT_NUM'], 2), device=self.dev)
    dr = self.th.clone(outputs)

    # We need to apply to blur operation on each "channel" (keypoint probability distribution) individually.
    #   In order to achieve this we reshape from (no_imgs, no_kpts, out_h, out_w) to (no_imgs * no_kpts, 1, out_h, out_w)
    #   then apply the blur then reshape again to (no_imgs, no_kpts, out_h, out_w)
    old_shape = dr.shape
    new_shape = (dr.shape[0] * dr.shape[1], 1, dr.shape[2], dr.shape[3])
    dr = dr.reshape(*new_shape)

    # Apply blur
    sigma = 0.3 * ((kernel - 1) * 0.5 - 1) + 0.8
    dr = self.tv.transforms.GaussianBlur(kernel, sigma)(dr)
    dr = dr.reshape(*old_shape)

    # Next we need to find the argmax for each keypoint "channel". In order to do this we need to reshape from
    #   (no_imgs, no_kpts, out_h, out_w) to (no_imgs * no_kpts, out_h * out_w) apply argmax and then unravel the index
    #   resulting in an index tensor with the shape (no_imgs, no_kpts, 2) that contains the x and y coord for each keypoint
    argmax_shape = (dr.shape[0] * dr.shape[1], dr.shape[2] * dr.shape[3])
    lbs = dr.reshape(*argmax_shape).argmax(dim=1)
    indexes = self.unravel_index(lbs, dr.shape[-2:]).reshape(*old_shape[:2], 2)

    # Mspn postprocess requires for some reason the make the maximums 0 and redo the argmax operation and find
    #   a new set of indexes `pindexes`
    lst_index_tensors = [self.th.Tensor([i for i in range(argmax_shape[0])]).to(self.dev).long(), lbs]

    zeros = self.th.zeros([1], device=self.dev)
    if self.cfg_fp16:
      zeros = zeros.half()
    dr = dr.reshape(*argmax_shape).index_put(
      lst_index_tensors,
      zeros
    ).reshape(*old_shape)

    lbs = dr.reshape(*argmax_shape).argmax(dim=1)
    pindexes = self.unravel_index(lbs, dr.shape[-2:]).reshape(*old_shape[:2], 2)

    pindexes -= indexes

    # Now we need to apply a shift to the indexes that have the second norm of their respective p indexes > 1e-2
    lns = ((pindexes[:, :, 0] ** 2 + pindexes[:, :, 1] ** 2) ** 0.5).unsqueeze(-1)
    indexes = indexes.float()
    indexes += (lns >= 1e-3) * (shifts[0] * pindexes / lns)

    # Next we rescale the value of the indexes to match the image shape
    indexes[:, :, 0] = self.th.clamp(indexes[:, :, 0], min=0, max=self.cfg_mspn_img_shape[0] - 1)
    indexes[:, :, 1] = self.th.clamp(indexes[:, :, 1], min=0, max=self.cfg_mspn_img_shape[1] - 1)
    indexes_scaled = indexes * 4 + 2

    kps[:, :, 0] = indexes_scaled[:, :, 1]
    kps[:, :, 1] = indexes_scaled[:, :, 0]

    # And least, we need to select the confidence score for each kpt by getting the value from the core maps
    #   First we cast the index to long
    int_indexes = (indexes.round() + 1e-9).long()

    #   Next we reshape the score map from (no_imgs, no_kpts, out_h, out_w) to (no_imgs * no_kpts, out_h * out_w) in
    #      order to use advanced indexing - there may be some solution without reshape, but I did not manage to find it
    scr_mps_rs = score_maps.reshape(
      (score_maps.shape[0] * score_maps.shape[1], score_maps.shape[2] * score_maps.shape[3])
    )
    #   Next we also need to reshape the indexes from (no_imgs, no_kpts, 2) to (no_imgs * no_kpts, 1)
    raveled_indexes = (int_indexes[:, :, 0] * score_maps.shape[3] + int_indexes[:, :, 1]).reshape(
      int_indexes.shape[0] * int_indexes.shape[1], 1
    )
    #   Finally, select the values and reshape the scores tensor to (no_imgs, no_kpts, 1)
    scores = scr_mps_rs[
      [i for i in range(raveled_indexes.shape[0])],
      raveled_indexes.squeeze(-1)
    ].reshape(score_maps.shape[0], score_maps.shape[1], 1)

    # Concat the keypoints and scores and normalize the positions to (0, 1)
    preds = self.th.cat([kps, scores], dim=2) / self.th.tensor([self.cfg_mspn_img_shape[1], self.cfg_mspn_img_shape[0], 1],
                                                               device=self.dev)

    if not warmup:
      self._stop_timer("pproc_img")

    return preds

  def mspn_stage(self, pred_nms, th_inputs):
    """
    input:
      Im1 (1 pers 1 dog 1 cat)
      Im2 (2 pers)
      Im3 (3 pers)
    output:
      [
       ["ON", None, None],
       ["OFF", "WRONG"]
       ["OFF", "ON", "ON"]
      ]


    :param pred_nms:
    :param th_inputs:
    :return:
      list(list<pred_nms>)

      list(
        im1: list(pred1, pred2 ...)
        im2: list(pred3, pred4, ...)
        im2: list(pred5, pred6, ...)
      )

      Iterate over pred_nms
      Create 2 lists :
        1. Each individual crop:
          Ex:
          [
            pers_img, dog_img, cat_img, pers_img, pers_img, pers_img, pers_img, pers_img,
          ]
        2. Crop image identity
          Ex:
          [
            0 0 0 1 1 2 2 2
          ]
      =====
      for each image crop tlbr from tensor; add to input list

      batch = th_resize_with_crop(input_list)
    """

    crop_imgs = []
    masks = []
    self._start_timer("crop")
    for i, pred in enumerate(pred_nms):
      pred_mask = (pred[:, 5] == self.class_names.index('person'))  # TODO: get `person` id
      masks.append(pred_mask.tolist())
      for i_crop in range(pred.shape[0]):
        if pred_mask[i_crop]:
          # in the future this check will disappear
          if not self.cfg_crop_original:
            crop_imgs.append(
              self.tv.transforms.functional.crop(
                th_inputs[i, :],
                left=max(pred[i_crop, 0].int(), 0),
                top=max(pred[i_crop, 1].int(), 0),
                width=max(pred[i_crop, 2].int(), 0) - max(pred[i_crop, 0].int(), 0) + 1,
                height=max(pred[i_crop, 3].int(), 0) - max(pred[i_crop, 1].int(), 0) + 1,
              )
            )
          else:
            crop_imgs.append(
              self._make_crop(
                frame=self.original_input_images[i],
                ltrb=pred[i_crop, :4].unsqueeze(0),
                original_shape=self._lst_original_shapes[i],
                return_offsets=False
              )
            )
        # endif
      # endfor
    # endfor
    self._stop_timer("crop")

    if len(crop_imgs) == 0:
      return [], []

    self._start_timer("resize")
    batch = self.th.cat([
      self.transforms(x.unsqueeze(0))
      for x in crop_imgs
    ])
    self._stop_timer("resize")

    self._start_timer("mspn_p_b{}_{}".format(len(batch), self.cfg_max_batch_second_stage))
    mspn_preds = self._batch_predict(
      prep_inputs=batch,
      model=self.mspn_model,
      batch_size=self.cfg_max_batch_second_stage,
      aggregate_batch_predict_callback=self._aggregate_second_stage_batch_predict
    )
    self._stop_timer("mspn_p_b{}_{}".format(len(batch), self.cfg_max_batch_second_stage))

    return mspn_preds, masks

  def compute_pred_indexes(self, masks):
    l_idxs = []
    k = 0
    cumsums = [self.np.cumsum(mask) for mask in masks]
    for i in range(len(masks)):
      l_idxs.append([
        cumsums[i][j] - 1 + k if masks[i][j] else None
        for j in range(len(masks[i]))
      ])
      k += cumsums[i][-1] if len(cumsums[i]) > 0 else 0
    # endfor
    return l_idxs

  def _second_stage_classifier(self, pred_nms, th_inputs):
    mspn_preds, masks = self.mspn_stage(pred_nms=pred_nms, th_inputs=th_inputs)
    if len(mspn_preds) == 0:
      return None
    results = []

    self._start_timer("pproc")
    move_preds = self._post_process_mspn(mspn_preds)
    self._stop_timer("pproc")

    self._start_timer("agg_res")
    move_preds_c = move_preds.cpu().numpy()
    l_idxs = self.compute_pred_indexes(masks)
    for idxs in l_idxs:
      results.append([
        self.np.round(move_preds_c[idx], 3) if idx is not None else None
        for idx in idxs
      ])
    self._stop_timer("agg_res")
    return results



import cv2
import numpy as np

# Build SIFT descriptors. Return keypoints and features
def get_features(frame):
  # The grayscale version of the image is considered for processing
  trainImg_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

  descriptor = None
  try:
    descriptor = cv2.SIFT_create()
  except:
    try:
      descriptor = cv2.xfeatures2d.SURF_create()
    except:
      pass
  
  if descriptor is None:
    raise ValueError('Could not calculate descriptor!')
  
  # Compute keypoints and features of the grayscale image
  kpsA, featuresA = descriptor.detectAndCompute(trainImg_gray, None)
  return (kpsA, featuresA)


# KNN for matching keypoints from anchor and test images
def match_keypoints_KNN(featuresA, featuresB, ratio):
  # Use BruteForce method to find matches and Euclidean norm
  bf = cv2.BFMatcher(cv2.NORM_L2, crossCheck=False)

  # __get_features() can return them None if image is obstructed
  if featuresA is not None and featuresB is not None:
    # Find the best 2 neighbours of each feature
    rawMatches = bf.knnMatch(featuresA, featuresB, 2)
    matches = []

    # Get only the matches that have a distance between them smaller than a threshold
    for mn in rawMatches:
      if len(mn) < 2:
        return None
      m, n = mn
      if m.distance < n.distance * ratio:
        matches.append(m)
    return matches
  # endif
  return None


# Build homography transformation in order to get the translation between images
def get_homography(kpsA, kpsB, matches, reprojThresh):
  # To float32
  kpsA = np.float32([kp.pt for kp in kpsA])
  kpsB = np.float32([kp.pt for kp in kpsB])

  # If homography transformation is possible
  if len(matches) > 4:
    # Get keypoints coordinates from matches
    ptsA = np.float32([kpsA[m.queryIdx] for m in matches])
    ptsB = np.float32([kpsB[m.trainIdx] for m in matches])

    # RANSAC algorithm is used to compute the best combinations between matches in order not to try all the possibilities
    (H, status) = cv2.findHomography(ptsA, ptsB, cv2.RANSAC,
                                     reprojThresh)

    return (matches, H, status)
  return None


from naeural_core import DecentrAIObject
from naeural_core import Logger
import abc

_CONFIG = {
  'VALIDATION_RULES' : {

  }
}

class BaseFileSystem(DecentrAIObject):
  CONFIG = _CONFIG
  def __init__(self, log: Logger, signature, config, **kwargs):
    self.signature = signature
    self.config = config
    super(BaseFileSystem, self).__init__(log=log, **kwargs)
    return

  def startup(self):
    super().startup()
    self.config_data = self.config
    return

  @abc.abstractmethod
  def upload(self, file_path, target_path):
    # TODO: maybe remove abstract and create & run a thread with _upload (abstract) target
    # also: put run_threaded FLAG !!!! default=False (plugins WILL be threads)
    # in main_loop use run_threaded=True (needed not to block main loop)
    raise NotImplementedError

  @abc.abstractmethod
  def download(self, uri, local_file_path):
    raise NotImplementedError

  def _http_download(self, url, local_file_path):
    # TODO: maybe review the case with multiple files downloaded
    saved_files, _ = self.log.maybe_download(
      url=url, fn=local_file_path, target=None,
      print_progress=False,
      force_download=True
    )

    if len(saved_files) == 1:
      return saved_files[0]

    return




import numpy as np
import abc
from naeural_core.local_libraries.nn.th.training.callbacks.base import TrainingCallbacks
from naeural_core.local_libraries.nn.th.training.callbacks.mixins import _ClassificationMetricsMixin

class ClassificationTrainingCallbacks(TrainingCallbacks, _ClassificationMetricsMixin, abc.ABC):
  def __init__(self, **kwargs):
    super(ClassificationTrainingCallbacks, self).__init__(**kwargs)
    return

  def get_class_names(self, dataset_info: dict = None, **kwargs) -> list[str]:
    """
    Get the class names from the dataset info
    Parameters
    ----------
    dataset_info : dict
      Dataset info dictionary
    kwargs : dict
      Additional keyword arguments

    Returns
    -------
    list[str]
      List of class names
    """
    return list(dataset_info['class_to_id'].keys())


  def _evaluate_callback(self, epoch: int, dataset_info: dict, y: np.ndarray, y_hat: np.ndarray, idx: np.ndarray, key:str = 'dev') -> dict:
    classes = self.get_class_names(dataset_info=dataset_info)
    dct_global_metrics = self.basic_metrics(y=y, y_hat=y_hat, classes=classes, key=key)


    new_best = self._owner.is_new_best(dct_global_metrics)
    if new_best or epoch % 10 == 0 or key == 'train': # TODO: Change key == 'train'
      dct_advanced_metrics = self.advanced_metrics(
        y=y, y_hat=y_hat, idx=idx, dataset_info=dataset_info, classes=classes, key=key
      )
      for k, dct_metrics in dct_advanced_metrics.items():
        self.P(".......... Category={} ..........".format(k))
        self.log_metrics(dct_metrics=dct_metrics, classes=classes, log_cm=new_best)
      #endfor
    #endfor

    self.log_metrics(dct_metrics=dct_global_metrics, classes=classes, log_cm=new_best)

    keys = list(dct_global_metrics.keys())
    lst_cm_keys = []
    for k in keys:
      if isinstance(dct_global_metrics[k], np.ndarray):
        lst_cm_keys.append(k)
    #endfor

    for k in lst_cm_keys:
      dct_global_metrics.pop(k)
    # endfor
    return dct_global_metrics

  def _test_callback(self, dataset_info: dict, y: np.ndarray, y_hat: np.ndarray, idx: np.ndarray) -> dict:
    classes = self.get_class_names(dataset_info=dataset_info)
    dct_metrics = self.basic_metrics(y=y, y_hat=y_hat, classes=classes, key='test')
    dct_advanced_metrics = self.advanced_metrics(
      y=y, y_hat=y_hat, idx=idx, dataset_info=dataset_info, classes=classes, key='test'
    )
    for k, subset_dct_metrics in dct_advanced_metrics.items():
      self.P(".......... Category={} ..........".format(k))
      self.log_metrics(dct_metrics=subset_dct_metrics, classes=classes, log_cm=True)
    #endfor
    self.log_metrics(dct_metrics=dct_metrics, classes=classes, log_cm=True)
    return dct_metrics

  def _dev_callback(self, dataset_info: dict, y: np.ndarray, y_hat: np.ndarray, idx: np.ndarray) -> dict:
    classes = self.get_class_names(dataset_info=dataset_info)
    dct_metrics = self.basic_metrics(y=y, y_hat=y_hat, classes=classes, key='dev')
    dct_advanced_metrics = self.advanced_metrics(
      y=y, y_hat=y_hat, idx=idx, dataset_info=dataset_info, classes=classes, key='dev'
    )
    for k, subset_dct_metrics in dct_advanced_metrics.items():
      self.P(".......... Category={} ..........".format(k))
      self.log_metrics(dct_metrics=subset_dct_metrics, classes=classes, log_cm=True)
    #endfor
    self.log_metrics(dct_metrics=dct_metrics, classes=classes, log_cm=True)
    return dct_metrics



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Run inference on images, videos, directories, streams, etc.

Usage:
    $ python path/to/detect.py --source path/to/img.jpg --weights yolov5s.pt --img 640
"""

import argparse
import sys
from pathlib import Path

import cv2
import numpy as np
import torch
import torch.backends.cudnn as cudnn

FILE = Path(__file__).resolve()
ROOT = FILE.parents[0]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
ROOT = ROOT.relative_to(Path.cwd())  # relative

from models.experimental import attempt_load
from utils.datasets import LoadImages, LoadStreams
from utils.general import apply_classifier, check_img_size, check_imshow, check_requirements, check_suffix, colorstr, \
    increment_path, non_max_suppression, print_args, save_one_box, scale_coords, set_logging, \
    strip_optimizer, xyxy2xywh
from utils.plots import Annotator, colors
from utils.torch_utils import load_classifier, select_device, time_sync


@torch.no_grad()
def run(weights=ROOT / 'yolov5s.pt',  # model.pt path(s)
        source=ROOT / 'data/images',  # file/dir/URL/glob, 0 for webcam
        imgsz=640,  # inference size (pixels)
        conf_thres=0.25,  # confidence threshold
        iou_thres=0.45,  # NMS IOU threshold
        max_det=1000,  # maximum detections per image
        device='',  # cuda device, i.e. 0 or 0,1,2,3 or cpu
        view_img=False,  # show results
        save_txt=False,  # save results to *.txt
        save_conf=False,  # save confidences in --save-txt labels
        save_crop=False,  # save cropped prediction boxes
        nosave=False,  # do not save images/videos
        classes=None,  # filter by class: --class 0, or --class 0 2 3
        agnostic_nms=False,  # class-agnostic NMS
        augment=False,  # augmented inference
        visualize=False,  # visualize features
        update=False,  # update all models
        project=ROOT / 'runs/detect',  # save results to project/name
        name='exp',  # save results to project/name
        exist_ok=False,  # existing project/name ok, do not increment
        line_thickness=3,  # bounding box thickness (pixels)
        hide_labels=False,  # hide labels
        hide_conf=False,  # hide confidences
        half=False,  # use FP16 half-precision inference
        ):
    source = str(source)
    save_img = not nosave and not source.endswith('.txt')  # save inference images
    webcam = source.isnumeric() or source.endswith('.txt') or source.lower().startswith(
        ('rtsp://', 'rtmp://', 'http://', 'https://'))

    # Directories
    save_dir = increment_path(Path(project) / name, exist_ok=exist_ok)  # increment run
    (save_dir / 'labels' if save_txt else save_dir).mkdir(parents=True, exist_ok=True)  # make dir

    # Initialize
    set_logging()
    device = select_device(device)
    half &= device.type != 'cpu'  # half precision only supported on CUDA

    # Load model
    w = weights[0] if isinstance(weights, list) else weights
    classify, suffix, suffixes = False, Path(w).suffix.lower(), ['.pt', '.onnx', '.tflite', '.pb', '']
    check_suffix(w, suffixes)  # check weights have acceptable suffix
    pt, onnx, tflite, pb, saved_model = (suffix == x for x in suffixes)  # backend booleans
    stride, names = 64, [f'class{i}' for i in range(1000)]  # assign defaults
    if pt:
        model = attempt_load(weights, map_location=device)  # load FP32 model
        stride = int(model.stride.max())  # model stride
        names = model.module.names if hasattr(model, 'module') else model.names  # get class names
        if half:
            model.half()  # to FP16
        if classify:  # second-stage classifier
            modelc = load_classifier(name='resnet50', n=2)  # initialize
            modelc.load_state_dict(torch.load('resnet50.pt', map_location=device)['model']).to(device).eval()
    elif onnx:
        check_requirements(('onnx', 'onnxruntime'))
        import onnxruntime
        session = onnxruntime.InferenceSession(w, None)
    else:  # TensorFlow models
        check_requirements(('tensorflow>=2.4.1',))
        import tensorflow as tf
        if pb:  # https://www.tensorflow.org/guide/migrate#a_graphpb_or_graphpbtxt
            def wrap_frozen_graph(gd, inputs, outputs):
                x = tf.compat.v1.wrap_function(lambda: tf.compat.v1.import_graph_def(gd, name=""), [])  # wrapped import
                return x.prune(tf.nest.map_structure(x.graph.as_graph_element, inputs),
                               tf.nest.map_structure(x.graph.as_graph_element, outputs))

            graph_def = tf.Graph().as_graph_def()
            graph_def.ParseFromString(open(w, 'rb').read())
            frozen_func = wrap_frozen_graph(gd=graph_def, inputs="x:0", outputs="Identity:0")
        elif saved_model:
            model = tf.keras.models.load_model(w)
        elif tflite:
            interpreter = tf.lite.Interpreter(model_path=w)  # load TFLite model
            interpreter.allocate_tensors()  # allocate
            input_details = interpreter.get_input_details()  # inputs
            output_details = interpreter.get_output_details()  # outputs
            int8 = input_details[0]['dtype'] == np.uint8  # is TFLite quantized uint8 model
    imgsz = check_img_size(imgsz, s=stride)  # check image size

    # Dataloader
    if webcam:
        view_img = check_imshow()
        cudnn.benchmark = True  # set True to speed up constant image size inference
        dataset = LoadStreams(source, img_size=imgsz, stride=stride, auto=pt)
        bs = len(dataset)  # batch_size
    else:
        dataset = LoadImages(source, img_size=imgsz, stride=stride, auto=pt)
        bs = 1  # batch_size
    vid_path, vid_writer = [None] * bs, [None] * bs

    # Run inference
    if pt and device.type != 'cpu':
        model(torch.zeros(1, 3, *imgsz).to(device).type_as(next(model.parameters())))  # run once
    dt, seen = [0.0, 0.0, 0.0], 0
    for path, img, im0s, vid_cap in dataset:
        t1 = time_sync()
        if onnx:
            img = img.astype('float32')
        else:
            img = torch.from_numpy(img).to(device)
            img = img.half() if half else img.float()  # uint8 to fp16/32
        img = img / 255.0  # 0 - 255 to 0.0 - 1.0
        if len(img.shape) == 3:
            img = img[None]  # expand for batch dim
        t2 = time_sync()
        dt[0] += t2 - t1

        # Inference
        if pt:
            visualize = increment_path(save_dir / Path(path).stem, mkdir=True) if visualize else False
            pred = model(img, augment=augment, visualize=visualize)[0]
        elif onnx:
            pred = torch.tensor(session.run([session.get_outputs()[0].name], {session.get_inputs()[0].name: img}))
        else:  # tensorflow model (tflite, pb, saved_model)
            imn = img.permute(0, 2, 3, 1).cpu().numpy()  # image in numpy
            if pb:
                pred = frozen_func(x=tf.constant(imn)).numpy()
            elif saved_model:
                pred = model(imn, training=False).numpy()
            elif tflite:
                if int8:
                    scale, zero_point = input_details[0]['quantization']
                    imn = (imn / scale + zero_point).astype(np.uint8)  # de-scale
                interpreter.set_tensor(input_details[0]['index'], imn)
                interpreter.invoke()
                pred = interpreter.get_tensor(output_details[0]['index'])
                if int8:
                    scale, zero_point = output_details[0]['quantization']
                    pred = (pred.astype(np.float32) - zero_point) * scale  # re-scale
            pred[..., 0] *= imgsz[1]  # x
            pred[..., 1] *= imgsz[0]  # y
            pred[..., 2] *= imgsz[1]  # w
            pred[..., 3] *= imgsz[0]  # h
            pred = torch.tensor(pred)
        t3 = time_sync()
        dt[1] += t3 - t2

        # NMS
        pred = non_max_suppression(pred, conf_thres, iou_thres, classes, agnostic_nms, max_det=max_det)
        dt[2] += time_sync() - t3

        # Second-stage classifier (optional)
        if classify:
            pred = apply_classifier(pred, modelc, img, im0s)

        # Process predictions
        for i, det in enumerate(pred):  # per image
            seen += 1
            if webcam:  # batch_size >= 1
                p, s, im0, frame = path[i], f'{i}: ', im0s[i].copy(), dataset.count
            else:
                p, s, im0, frame = path, '', im0s.copy(), getattr(dataset, 'frame', 0)

            p = Path(p)  # to Path
            save_path = str(save_dir / p.name)  # img.jpg
            txt_path = str(save_dir / 'labels' / p.stem) + ('' if dataset.mode == 'image' else f'_{frame}')  # img.txt
            s += '%gx%g ' % img.shape[2:]  # print string
            gn = torch.tensor(im0.shape)[[1, 0, 1, 0]]  # normalization gain whwh
            imc = im0.copy() if save_crop else im0  # for save_crop
            annotator = Annotator(im0, line_width=line_thickness, example=str(names))
            if len(det):
                # Rescale boxes from img_size to im0 size
                det[:, :4] = scale_coords(img.shape[2:], det[:, :4], im0.shape).round()

                # Print results
                for c in det[:, -1].unique():
                    n = (det[:, -1] == c).sum()  # detections per class
                    s += f"{n} {names[int(c)]}{'s' * (n > 1)}, "  # add to string

                # Write results
                for *xyxy, conf, cls in reversed(det):
                    if save_txt:  # Write to file
                        xywh = (xyxy2xywh(torch.tensor(xyxy).view(1, 4)) / gn).view(-1).tolist()  # normalized xywh
                        line = (cls, *xywh, conf) if save_conf else (cls, *xywh)  # label format
                        with open(txt_path + '.txt', 'a') as f:
                            f.write(('%g ' * len(line)).rstrip() % line + '\n')

                    if save_img or save_crop or view_img:  # Add bbox to image
                        c = int(cls)  # integer class
                        label = None if hide_labels else (names[c] if hide_conf else f'{names[c]} {conf:.2f}')
                        annotator.box_label(xyxy, label, color=colors(c, True))
                        if save_crop:
                            save_one_box(xyxy, imc, file=save_dir / 'crops' / names[c] / f'{p.stem}.jpg', BGR=True)

            # Print time (inference-only)
            print(f'{s}Done. ({t3 - t2:.3f}s)')

            # Stream results
            im0 = annotator.result()
            if view_img:
                cv2.imshow(str(p), im0)
                cv2.waitKey(1)  # 1 millisecond

            # Save results (image with detections)
            if save_img:
                if dataset.mode == 'image':
                    cv2.imwrite(save_path, im0)
                else:  # 'video' or 'stream'
                    if vid_path[i] != save_path:  # new video
                        vid_path[i] = save_path
                        if isinstance(vid_writer[i], cv2.VideoWriter):
                            vid_writer[i].release()  # release previous video writer
                        if vid_cap:  # video
                            fps = vid_cap.get(cv2.CAP_PROP_FPS)
                            w = int(vid_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                            h = int(vid_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                        else:  # stream
                            fps, w, h = 30, im0.shape[1], im0.shape[0]
                            save_path += '.mp4'
                        vid_writer[i] = cv2.VideoWriter(save_path, cv2.VideoWriter_fourcc(*'mp4v'), fps, (w, h))
                    vid_writer[i].write(im0)

    # Print results
    t = tuple(x / seen * 1E3 for x in dt)  # speeds per image
    print(f'Speed: %.1fms pre-process, %.1fms inference, %.1fms NMS per image at shape {(1, 3, *imgsz)}' % t)
    if save_txt or save_img:
        s = f"\n{len(list(save_dir.glob('labels/*.txt')))} labels saved to {save_dir / 'labels'}" if save_txt else ''
        print(f"Results saved to {colorstr('bold', save_dir)}{s}")
    if update:
        strip_optimizer(weights)  # update model (to fix SourceChangeWarning)


def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument('--weights', nargs='+', type=str, default=ROOT / 'yolov5l6.pt', help='model path(s)')
    parser.add_argument('--source', type=str, default=ROOT / 'data/images', help='file/dir/URL/glob, 0 for webcam')
    parser.add_argument('--imgsz', '--img', '--img-size', nargs='+', type=int, default=[1280], help='inference size h,w')
    parser.add_argument('--conf-thres', type=float, default=0.25, help='confidence threshold')
    parser.add_argument('--iou-thres', type=float, default=0.45, help='NMS IoU threshold')
    parser.add_argument('--max-det', type=int, default=1000, help='maximum detections per image')
    parser.add_argument('--device', default='0', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
    parser.add_argument('--view-img', action='store_true', help='show results')
    parser.add_argument('--save-txt', action='store_true', help='save results to *.txt')
    parser.add_argument('--save-conf', action='store_true', help='save confidences in --save-txt labels')
    parser.add_argument('--save-crop', action='store_true', help='save cropped prediction boxes')
    parser.add_argument('--nosave', action='store_true', help='do not save images/videos')
    parser.add_argument('--classes', nargs='+', type=int, help='filter by class: --classes 0, or --classes 0 2 3')
    parser.add_argument('--agnostic-nms', action='store_true', help='class-agnostic NMS')
    parser.add_argument('--augment', action='store_true', help='augmented inference')
    parser.add_argument('--visualize', action='store_true', help='visualize features')
    parser.add_argument('--update', action='store_true', help='update all models')
    parser.add_argument('--project', default=ROOT / 'runs/detect', help='save results to project/name')
    parser.add_argument('--name', default='exp', help='save results to project/name')
    parser.add_argument('--exist-ok', action='store_true', help='existing project/name ok, do not increment')
    parser.add_argument('--line-thickness', default=3, type=int, help='bounding box thickness (pixels)')
    parser.add_argument('--hide-labels', default=False, action='store_true', help='hide labels')
    parser.add_argument('--hide-conf', default=False, action='store_true', help='hide confidences')
    parser.add_argument('--half', action='store_true', help='use FP16 half-precision inference')
    opt = parser.parse_args()
    opt.imgsz *= 2 if len(opt.imgsz) == 1 else 1  # expand
    print_args(FILE.stem, opt)
    return opt


def main(opt):
    # check_requirements(exclude=('tensorboard', 'thop'))
    run(**vars(opt))


if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



import sys
import torch as th
import torchvision as tv
import time
import json
import os
import gc
import numpy as np
import abc

from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad
from ratio1 import load_dotenv


load_dotenv()


class BaseScripter(metaclass=abc.ABCMeta):
  """
  This class will be used in order to convert already existent pytorch models to
  various formats via torch tracing.
  """

  def __init__(
    self, log, model, model_name, input_shape,
    model_config=None, preprocess_method=None,
    matching_method=None,
    predict_method=None, use_fp16=False,
    use_amp=False,
    gold_model=None
  ):
    """
    Parameters
    ----------
    log - Logger, swiss knife object used in all DecentrAI
    model - torch.nn.Module, the instantiated model to be converted
    model_name - str, name of the model
    input_shape - list, expected input shape by the model
    model_config - dict, configuration of the converted model
    preprocess_method - method, method of preprocessing the input in case it is needed
    - for the default value of this see self.preprocess_callback
    matching_method - method, method for checking if the output of the model matches the
      output of the traced model
    - for the default value of this see self.matching_callback
    predict_method - method, method of running inferences given model and inputs
    - for the default value of this see self.default_predict
    use_fp16 - bool, whether to use fp16 when tracing a model
    use_amp - bool, whether to use amp when tracing a model
    gold_model - Torchscript model, if not None we will use this for validation
    """
    self.model = model
    self.model_name = model_name
    self.model_config = {} if model_config is None else model_config
    self.log = log
    self.input_shape = input_shape
    self.preprocess_method = preprocess_method if preprocess_method is not None else self.preprocess_callback
    self.matching_method = matching_method if matching_method is not None else self.matching_callback
    self.predict_method = predict_method if predict_method is not None else self.default_predict
    self.use_amp = use_amp
    self.use_fp16 = use_fp16
    self.gold_model = gold_model
    self.model.is_gold = False
    if self.use_fp16:
      self.model.half()
      if self.gold_model is not None:
        self.gold_model.half()
        self.gold_model.is_gold = True
    return

  def clear_cache(self):
    gc.collect()
    th.cuda.empty_cache()
    return

  def start_timer(self, name):
    self.log.start_timer(name, section='tracing')
    return

  def stop_timer(self, name):
    self.log.stop_timer(name, section='tracing')
    return

  def default_predict(self, model, inputs):
    return model(inputs)

  def __predict(self, model, inputs):
    if self.use_amp:
      with th.cuda.amp.autocast(enabled=self.use_amp):
        with th.no_grad():
          return self.predict_method(model, inputs)
    else:
      with th.no_grad():
        return self.predict_method(model, inputs)
    # endif use_amp

  def preprocess_callback(self, inputs, device='cpu', normalize=True, **kwargs):
    """
    This default callback will deal with resizing images, hence the majority of the models used in DecentrAI are
    computer vision models at the moment.
    Parameters
    ----------
    inputs - list, list of inputs on which to trace/test the model
    device - str or torch.device, device on which the input will be put

    Returns
    -------
    res - the data ready to enter the model
    """
    h, w = self.input_shape[:2]
    self.log.P("  Resizing from {} to {}".format([x.shape for x in inputs], (h, w)))
    results = th_resize_with_pad(
      img=inputs,
      h=h,
      w=w,
      device=device,
      normalize=normalize,
      return_original=False,
      half=self.use_fp16
    )
    if len(results) < 3:
      prep_inputs, lst_original_shapes = results
    else:
      prep_inputs, lst_original_shapes, lst_original_images = results
    return prep_inputs

  def convert_to_batch_size(self, inputs, batch_size):
    if hasattr(inputs, 'shape') and len(inputs.shape) == len(self.input_shape):
      # here inputs will be a single input
      inputs = [inputs for _ in range(batch_size)]
    else:
      # here inputs will be a list of inputs
      if hasattr(inputs, '__len__') and len(inputs) != batch_size:
        inputs = [inputs[i % len(inputs)] for i in range(batch_size)]
      # endif inputs of different length than the batch size needed
    # endif inputs is a single input
    return inputs

  def model_timing(self, model, model_name, inputs, nr_warmups=20, nr_tests=20, **kwargs):
    """
    Method for timing the speed of a certain model on given inputs.
    Parameters
    ----------
    model - torch.nn.Module, the model to be checked
    model_name - str, name of the timed model
    inputs - any, input data on which the model will be traced
    nr_warmups - int, how many inferences to make for warm up
    nr_test - int, how many inferences to make for timings

    Returns
    -------
    res - float, average time of inference for a batch
    """
    self.log.restart_timer(model_name)
    # warmup
    self.log.P(f"  Warming up ({nr_warmups} inferences)...")
    for _ in range(nr_warmups):
      print('.', flush=True, end='')
      preds = self.__predict(model, inputs)
    print('')

    # timing
    self.log.P(f"  Predicting ({nr_tests} inferences)...")
    for _ in range(nr_tests):
      print('.', flush=True, end='')
      self.start_timer(model_name)
      preds = self.__predict(model, inputs)
      self.stop_timer(model_name)
    print('')
    return self.log.get_timer_mean(model_name)

  def maybe_numpy(self, x):
    if isinstance(x, th.Tensor):
      return x.detach().cpu().numpy()
    return x

  def matching_callback(self, output1, output2, atol=0.0001, **kwargs):
    """
    Method for validating that 2 outputs are matching
    Parameters
    ----------
    output1 - any
    output2 - any
    atol - float, absolute tolerance for the comparison

    Returns
    -------
    True if outputs are matching, False otherwise
    """
    if isinstance(output1, tuple) and len(output1) == 1:
      output1 = output1[0]
    if isinstance(output2, tuple) and len(output2) == 1:
      output2 = output2[0]

    if type(output1) != type(output2):
      self.log.P('TYPE IS DIFFERENT! {} {} '.format(type(output1), type(output2)))
      return False
    if isinstance(output1, (list, tuple)):
      return len(output1) == len(output2) and all([
        np.allclose(self.maybe_numpy(output1[i]), self.maybe_numpy(output2[i]), atol=atol)
        for i in range(len(output1))]
      )
    if isinstance(output1, th.Tensor):
      output1 = output1.detach().cpu().numpy()
      output2 = output2.detach().cpu().numpy()

    if (isinstance(output1, np.ndarray) and np.any(output1 != output2)) or (not isinstance(output1, np.ndarray) and output1 != output2):
      try:
        self.log.P('Absolute error is {}'.format(np.max(np.abs(output1 - output2))))
        self.log.P('Mean error is {}'.format(np.average(np.abs(output1 - output2))))
        self.log.P('Std error is {}'.format(np.std(np.abs(output1 - output2))))
        return np.allclose(output1, output2, atol=atol)
      except Exception as e:
        self.log.P(f'Error! {e}')
        return False
    # endif outputs are different
    return True

  @abc.abstractmethod
  def load(self, fn_path, device, batch_size, use_fp16):
    """
    Loads the traced model from disk to memory, possibly converting it to float16.

    Parameters
    ----------
    fn_path: str, path on disk to the traced model

    device: th.device, the torch device to load the model to

    batch_size: maximum supported batch size for inference

    use_fp16: if True, converts the loaded model to f16

    Returns
    -------
    Tuple[model, dict] - a tuple containing the model and the model
      configuration
    """
    pass

  @abc.abstractmethod
  def convert(self, inputs, config, model_fn):
    """
    Traces the pytorch model of the scripter, saving it to disk.

    Parameters
    ----------

    inputs - data for tracing (FIXME: what's the exact type??)

    config : dict, model config

    model_fn : str, path where the model should be saved on disk

    Returns
    -------
    Union[ModelBackendWrapper, torch.jit.ScriptModule] - the converted model,
    loaded from disk
    """
    pass

  def test(
      self, ts_path, model, inputs, batch_size=2, device='cpu',
      nr_warmups=20, nr_tests=20, skip_preprocess=False,
      use_fp16=False, **kwargs
  ):
    """
    Method for testing the traced model at the same time with the model in order to
    validate that the outputs are the same and to check the speed difference between the 2.
    Parameters
    ----------
    ts_path - str, path to the traced model
    model - torch.nn.Module, the model to be checked
    inputs - any, input data on which the model will be traced
    batch_size - int, Warning: the batch_size for testing and the batch
    size on which the model was traced should be different in order to
    further check that the model was traced correctly
    device - str or torch.device
    skip_preprocess - bool, whether to skip the preprocessing of the input
    nr_warmups - int, how many inferences to make for warm up
    nr_test - int, how many inferences to make for timings
    use_fp16 - bool, whether to convert both the model and the traced model to
    float16 before testing

    Returns
    -------
    res - (ok, timings), where:
    ok - bool, True if the outputs coincide, False otherwise
    timings - dict, the timers of both the model and the traced version
    """
    str_prefix = f'[bs={batch_size}]'
    self.log.P(f"{str_prefix}Testing the graph from {ts_path} for {self.model_name} with bs={batch_size} on device {device} with use_fp16={use_fp16}")
    ts_model, config = self.load(ts_path, device, batch_size=batch_size, use_fp16=use_fp16)
    ts_input_shape = config.get('input_shape')
    assert list(ts_input_shape) == list(self.input_shape), \
        f'Error! The input shape of the model and the .ths do not coincide! ' \
        f'The specified shape for the model was {self.input_shape}, while the ' \
        f'input shape in the .ths config is {ts_input_shape}'

    if not skip_preprocess:
      self.log.P(f'{str_prefix}Preprocessing...')
      inputs = self.convert_to_batch_size(inputs=inputs, batch_size=batch_size)
      inputs = self.preprocess_method(inputs=inputs, device=device, **kwargs)
    else:
      self.log.P(f'{str_prefix}Skipping preprocessing')
    # endif skip_preprocess

    if model.is_gold:
      self.log.P("Running with gold model")
    else:
      self.log.P("Running without gold model")

    suffix_str = ""
    if use_fp16:
      model.half()
      inputs = inputs.half()
      suffix_str = "[FP16]"
    # endif use_fp16

    self.log.P(f'{str_prefix}Inferences...')
    model_output = self.__predict(model, inputs)
    ts_output = self.__predict(ts_model, inputs)

    self.log.P(f'{str_prefix}Validating...')
    ok = self.matching_method(model_output, ts_output, **kwargs)

    self.log.P(f'{str_prefix}Timing...')
    model_time = self.model_timing(
      model=model, model_name=self.model_name, inputs=inputs,
      nr_warmups=nr_warmups, nr_tests=nr_tests, **kwargs
    )
    self.clear_cache()
    ts_time = self.model_timing(
      model=ts_model, model_name=self.model_name + '_ts', inputs=inputs,
      nr_warmups=nr_warmups, nr_tests=nr_tests, **kwargs
    )
    self.clear_cache()
    timing_dict = {
      f'{self.model_name}_python': model_time,
      f'{self.model_name}_ts': ts_time,
      f'{self.model_name}_python_per_input': model_time / batch_size,
      f'{self.model_name}_ts_per_input': ts_time / batch_size,
    }
    self.log.P(f'{str_prefix}Model outputs are{"" if ok else " not"} matching!!{suffix_str}', color='g' if ok else 'r')
    total_gain = model_time - ts_time
    model_time_per_input = model_time / batch_size
    ts_time_per_input = ts_time / batch_size
    self.log.P(f'{str_prefix}Time for normal model: {model_time:.05f}s [{model_time_per_input:.05f} per input]{suffix_str}')
    self.log.P(f'{str_prefix}Time for traced model: {ts_time:.05f}s [{ts_time_per_input:.05f} per input]{suffix_str}')
    self.log.P(
      f'{str_prefix}Speed gain of {total_gain:.05f}s per batch and {total_gain / batch_size:.05f}s per input! ({total_gain/model_time*100:.02f}%){suffix_str}',
      color='g' if total_gain > 0 else 'r'
    )
    return ok, timing_dict

  def generate(
      self, inputs, batch_size=1, device='cpu', to_test=False,
      nr_warmups=20, nr_tests=20, no_grad_tracing=True,
      test_batch_size=None, test_fp16=False, **kwargs
  ):
    """
    Method for generating the traced model on a given device, batch size and inputs.
    Parameters
    ----------
    inputs - any, input data on which the model will be traced
    batch_size - int
    device - str or torch.device
    to_test - bool, whether to test the model after tracing
    nr_warmups - int, how many inferences to make for warm up, relevant only if to_test=True
    nr_test - int, how many inferences to make for timings, relevant only if to_test=True
    no_grad_tracing - bool, whether to apply th.no_grad() when tracing the model
    test_batch_size - int or None, the batch size on which to test the model
    - if None it will be 2 x batch_size
    - relevant only if to_test=True
    test_fp16 - bool, whether to test the model in fp16

    Returns
    -------
    res - path of the saved traced model
    """
    self.log.P(f"Generating traced model for {self.model_name} on {device} with batch_size={batch_size} "
               f"and the following config:")
    self.log.P(f"{self.model_config}")
    config = {
      **self.model_config,
      'input_shape': self.input_shape,
      'python': sys.version.split()[0],
      'torch': th.__version__,
      'torchvision': tv.__version__,
      'device': device,
      'optimize': False,
      'date': time.strftime('%Y%m%d', time.localtime(time.time())),
      'model': self.model_name,
    }
    original_inputs = inputs
    inputs = self.convert_to_batch_size(inputs=original_inputs, batch_size=batch_size)
    inputs = self.preprocess_method(inputs=inputs, device=device, **kwargs)

    self.log.P("  Scripting...")
    self.model.to(device)

    save_dir = os.path.join(self.log.get_models_folder(), 'traces')
    os.makedirs(save_dir, exist_ok=True)
    model_fn = self.model_name + f'{"_fp16" if self.use_fp16 else ""}_bs{batch_size}'
    model_fn = model_fn + self.extension
    fn = os.path.join(save_dir, model_fn)

    traced_model = self.convert(inputs, config, fn)

    self.log.P("  Forwarding using traced...")
    output = self.__predict(traced_model, inputs)

    loaded_model, config = self.load(fn, device=device, batch_size=batch_size, use_fp16=test_fp16)
    self.log.P("  Loaded config with {}".format(list(config.keys())))

    prep_inputs_test = self.convert_to_batch_size(inputs=original_inputs, batch_size=2 * batch_size)
    prep_inputs_test = self.preprocess_method(inputs=prep_inputs_test, device=device, **kwargs)

    self.log.P("  Running forward...")
    preds = self.__predict(loaded_model, prep_inputs_test)
    self.log.P(f"  Done running forward. Ouput:\n{preds}")

    if to_test:
      test_batch_size = 2 * batch_size if test_batch_size is None else test_batch_size
      if self.gold_model is not None:
        self.gold_model.to(device)

      self.log.P('Starting validation phase...')
      if self.gold_model is None:
        self.log.P("Starting test without gold model")
      else:
        self.log.P("Starting test WITH gold model")

      test_kwargs = {
        'ts_path': fn,
        'model': self.model if self.gold_model is None else self.gold_model,
        'inputs': prep_inputs_test,
        'batch_size': test_batch_size,
        'device': device,
        'nr_warmups': nr_warmups,
        'nr_tests': nr_tests,
        'skip_preprocess': True,
        **kwargs
      }
      self.test(**test_kwargs)
      if test_fp16:
        self.test(use_fp16=True, **test_kwargs)
    # endif to_test

    return fn




# -*- coding: utf-8 -*-
"""

Pseudo-code for business plugin threading design.

1. Plugin modification adds deque of upstream data and a loop method (thread target)
2. Manager justs adds data to plugin queues
Now the hard part:
3. Communicator manager on the other hand must be a thread that continously checks for available payloads
  thus the payloads must be

"""


# ALMOST DONE
# - TODO - individual queues must be monitored by the AppMonitor and and alerts should be sent if any plugin is too lazy
# - TODO - implement _on_idle in jobs plugins (blur, object-detector-tracking)
class Plugin():

    # def add_inputs(self, inputs):
    #   self.upstream_data_deque.append(inputs)
    #   return

    def _on_data(self):
        # default behaviour execution chain but can be overwritten
        self.execution_chain()
        return

    def _on_idle(self):
        # can be overwritten
        return

    def _on_init(self):
        # can be overwritten
        return

    def _on_close(self):
        # can be overwritten
        return

    def plugin_loop(self):
        # thread control
        self.done = False
        # one timer
        try:
            self._on_init()
            self.P("Thread ... initialized")
            # loop
            while not self.done:
                # upstream data is appended to a deque
                if len(self.upstream_data_deque) > 0:
                    # if data then call "On Data"
                    # IMPORTANT: individual queues must be monitored by the AppMonitor
                    #            and allerts should be sent if any plugin is too lazy
                    self.inputs = self.upstream_data_deque.popleft()
                    self.start_timer('on_data')
                    self._on_data()
                    self.stop_timer('on_data')
                else:
                    # call "On Idle"
                    self.start_timer('on_idle')
                    self._on_idle()
                    self.stop_timer('on_idle')

                # now we may have some output
                payload = self.get_payload_after_exec()
                commands = self.get_commands_after_exec()
                # if valid then add to output queues
                if payload is not None:
                    self.payloads_queue.append(payload)
                if commands is not None:
                    self.commands_queue.append(commands)

                # plugin must be FASTER than main loop
                # cfg_plugin_loop_resolution = MAX(get(_CONFIG, 20), get(MAIN_LOOP))
                sleep(1 / self.cfg_plugin_loop_resolution)

            self._on_close()
        except:
            self.P("Thread exception")

        self.P("Thread ... end")
        return

# DONE


def execute_non_threaded(self):
    self.inputs = self.upstream_data_deque.popleft()
    if self.inputs is not None:
        self._on_data()
    else:
        self._on_idle()

    payload = self.get_payload_after_exec()
    commands = self.get_commands_after_exec()
    # if valid then add to output queues
    if payload is not None:
        self.payloads_queue.append(payload)
    if commands is not None:
        self.commands_queue.append(commands)

# DONE manager stuff

# DONE


def _check_instances(self,):
    # ...
    # at instantiezation
    self.comm_shared_memory['payloads'][instance_hash] = deque()
    self.comm_shared_memory['commands'][instance_hash] = deque()
    plugin.payloads_queue = self.dct_payloads_queues[instance_hash]
    plugin.commands_queue = self.dct_commands_queues[instance_hash]
    # now PluginsManager.comm_shared_memory for the new plugin is available for external .popleft()
    # ...

    plugin = _cls_def(
        log=self.log,
        global_shmem=self.shmem,
        stream_id=stream_name,
        signature=signature,
        default_config=_config_dict,
        upstream_config=upstream_config,
        shared_data=self._shared_data,
        initiator_id=initiator_id,
        session_id=session_id,
        threaded_execution_chain=self._run_on_threads,
    )

    self.P("New `{}` added in the loop.".format(repr(plugin)), color='g')

    self._dct_current_instances[instance_hash] = plugin
    if self._run_on_threads:
        plugin._thread = Thread(target=plugin.plugin_loop, daemon=True)
        plugin._thread.start()

# DONE


def execute_all_plugins(self, dct_plugins_inputs, payload_type=None):
    valid_payload_types = ['INSTANCE', 'STREAM', 'BOX']

    if payload_type is None:
        payload_type = 'STREAM'
    # endif

    if payload_type not in valid_payload_types:
        self.P("ERROR! Payload type {} not valid. We take the default one - {}".format(
            payload_type, valid_payload_types[0]))
        payload_type = valid_payload_types[0]
    # endif

    for instance_hash, inputs in dct_plugins_inputs.items():
        plugin = self.get_subaltern(instance_hash)
        plugin.add_inputs(inputs)
        if not self._run_on_threads:
            plugin.execute_non_threaded()

    # and ze ze zeet
    return

# DONE


def orchestrator_comm_thread(self):
    while not self.__done:
        return_code, status = self.communicate_recv_and_handle()
        payloads = []
        commands = []
        payload_instances = list(self.__comm_shared_memory['payloads'].keys())
        for instance_hash in payload_instances:
            while (self.__comm_shared_memory['payloads'][instance_hash]) > 0:
                payloads.append(
                    self.__comm_shared_memory['payloads'][instance_hash].popleft())

        commands_instances = list(self.__comm_shared_memory['commands'].keys())
        for instance_hash in commands_instances:
            while (self.__comm_shared_memory['commands'][instance_hash]) > 0:
                commands.append(
                    self.__comm_shared_memory['commands'][instance_hash].popleft())

        self.communicate_send(
            payloads=payloads,
            commands=commands,
            status=status
        )  # includes a _maybe_send_heartbeat()
        # orchestrator comm thread must be faster than main loop
        sleep(1 / (self.cfg_main_loop_resolution + DELTA))  # maybe DELTA = 1
    # now force any remaining buffer



from naeural_core import DecentrAIObject
import pandas as pd
import time as tm

class AbstractConnector(DecentrAIObject):
  def __init__(self, **kwargs):
    self.reader = None
    self.df_item_mapping = None

    super().__init__(prefix_log='[CONN]', **kwargs)
    return

  def set_reader(self):
    self.D("Setting reader ...")
    return

  def chunk_generator(self):
    self.D("Chunck generator.")
    self.set_reader()
    for df_chunk in self.reader:
      d1, d2 = df_chunk.shape
      self.D("Chunk events: {} / Chunk columns: {}".format(d1, d2))
      yield df_chunk


class CSVConnector(AbstractConnector):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    self.parse_dates = None
    return

  def set_data_connector_params(self,
                                dataset=None,
                                fn_data=None,
                                absolute_path_fn_data=None,
                                chunksize=None,
                                parse_dates=None):
    self._parse_config_data(
      dataset,
      fn_data=fn_data,
      absolute_path_fn_data=absolute_path_fn_data,
      chunksize=chunksize
    )
    self.parse_dates = parse_dates
    return

  def set_item_mapping(self, fn_nomenclator=None):
    if fn_nomenclator is not None:
      if fn_nomenclator != '':
        self.D("Setting item mapping ...")
        self.df_item_mapping = self.log.load_dataframe(fn_nomenclator)
    return

  def set_reader(self):
    assert self.fn_data is not None
    super().set_reader()

    if type(self.fn_data) is not list:
      if self.absolute_path_fn_data:
        fn = self.fn_data
      else:
        fn = self.log.get_data_file(self.fn_data)

      self.D("Created csv connector based on '{}' with chunksize={}"
             .format(fn, self.chunksize))
      self.reader = pd.read_csv(fn, iterator=True, chunksize=self.chunksize,
                                parse_dates=self.parse_dates)
    else:
      if not self.absolute_path_fn_data:
        files = list(map(lambda x: self.log.get_data_file(x), self.fn_data))

      self.D("Created csv connector based on multiple files '{}' with chunksize={}"
             .format(files, self.chunksize))

      self.reader = self.chunk_from_files(*files)
    return

  def _chunk_from_files(self, *files):
    for fn in files:
      for chunk in pd.read_csv(fn, iterator=True, chunksize=self.chunksize,
                               parse_dates=self.parse_dates):
        yield chunk

class DFConnector(AbstractConnector):
  def __init__(self, **kwargs):
    self.df = None
    super(DFConnector, self).__init__(**kwargs)
    return

  def set_data_connector_params(self, df):
    self.df = df
    return

  def _pseudo_gen(self):
    yield self.df

  def set_reader(self):
    super().set_reader()
    self.reader = self._pseudo_gen()
    return


class ODBCConnector(AbstractConnector):
  def __init__(self, **kwargs):
    self.sql_query = "SELECT * FROM {};"
    super().__init__(**kwargs)
    return

  def set_data_connector_params(self, dataset=None, table_data=None,
                                odbc_driver=None, server=None, port=None,
                                database=None, uid=None, password=None,
                                chunksize=None, nr_conn_retry=15):
    self._parse_config_data(
      dataset,
      table_data=table_data,
      odbc_driver=odbc_driver,
      server=server,
      port=port,
      database=database,
      uid=uid,
      password=password,
      chunksize=chunksize,
      nr_conn_retry=nr_conn_retry
    )

    for i, x in enumerate([self.table_data, self.odbc_driver,
                           self.server, self.port,
                           self.database, self.uid, self.password,
                           self.nr_conn_retry]):
      assert x is not None, 'Variable #{} is None'.format(i)

    str_conn = "DRIVER={}; ".format(self.odbc_driver)
    str_conn += "SERVER={}; ".format(self.server)
    str_conn += "PORT={}; ".format(self.port)
    str_conn += "DATABASE={}; ".format(self.database)
    str_conn += "UID={}; ".format(self.uid)
    str_conn += "PWD={};".format(self.password)

    count = 0
    connected = False

    while count < self.nr_conn_retry and connected == False:
      count += 1
      try:
        import pyodbc
        self.D("ODBC Conn try {}: {}...".format(count, str_conn))
        self.cnxn = pyodbc.connect(str_conn)
        self.D("Connection created.")
        connected = True
      except Exception as err:
        self.D("FAILED ODBC Conn on retry: {}!".format(count))
        self.D("ERROR: {}".format(err))
        tm.sleep(0.5)

    return

  def set_item_mapping(self, table_nomenclator=None):
    self.D("Setting item mapping ...")
    self.df_item_mapping = pd.read_sql(sql=self.sql_query.format(table_nomenclator),
                                       con=self.cnxn)
    return

  def set_reader(self):
    super().set_reader()

    sql_query = "SELECT COUNT(*) FROM {};".format(self.table_data)
    cursor = self.cnxn.cursor()
    cursor.execute(sql_query)
    nr_rows = cursor.fetchone()[0]

    if self.chunksize is None:
      self.chunksize = nr_rows

    self.reader = pd.read_sql(sql=self.sql_query.format(self.table_data),
                              con=self.cnxn,
                              chunksize=self.chunksize)
    self.D("Created ODBC connector based on '{}.{}' with chunksize={}"
           .format(self.database, self.table_data, self.chunksize))
    return


if __name__ == '__main__':
  from naeural_core import Logger

  log = Logger(lib_name='C',
               config_file='radiography/config.txt',
               TF_KERAS=False)
  c = ODBCConnector(log)
  c.set_data_connector_params()
  gen = c.chunk_generator()


"""
This serving process is an experimental one.
It is designed for testing several Yolo models with different input sizes.
"""
from naeural_core.serving.default_inference.th_yf_base import YfBase as ParentServingProcess

FILENAMES = [
  r"C:\repos\edge-node\core\xperimental\th_y8\20240304_y8l_640x1152_nms.ths",
  r"C:\repos\edge-node\core\xperimental\th_y8\20240304_y8l_768x768_nms.ths",
  r"C:\repos\edge-node\core\xperimental\th_y8\20240304_y8l_896x896_nms.ths",
  r"C:\repos\edge-node\core\xperimental\th_y8\20240304_y8s_576x1024_nms.ths",
  r"C:\repos\edge-node\core\xperimental\th_y8\20240304_y8s_640x1152_nms.ths",
  r"C:\repos\edge-node\core\xperimental\th_y8\20240304_y8s_768x768_nms.ths",
]

_CONFIG = {
  **ParentServingProcess.CONFIG,

  "MODEL_WEIGHTS_FILENAME": FILENAMES[4],
  "MODEL_WEIGHTS_FILENAME_DEBUG": "20230723_y8l_nms_top6.ths",

  "URL": "minio:Y8/20230723_y8l_nms.ths",
  "URL_DEBUG": "minio:Y8/20230723_y8l_nms_top6.ths",

  'VALIDATION_RULES': {
    **ParentServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

DEBUG_COVERED_SERVERS = False

__VER__ = '0.2.0.0'


class ThYfXp(ParentServingProcess):
  @property
  def cfg_input_size(self):
    fn = self.get_model_weights_filename
    cfg = self.graph_config[fn]
    imgsz = cfg['imgsz']
    return imgsz



from naeural_core.business.base.cv_plugin_executor import CVPluginExecutor as BasePlugin


_CONFIG = {
  **BasePlugin.CONFIG,
  'PROCESS_DELAY': 0,

  "AI_ENGINE": "lowres_general_detector",
  "COLOR_TAGGING": True,
  "DEBUG_DETECTIONS": False,

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES']
  },
}


class ObjectDetectorPlugin(BasePlugin):
  """
  Demo plugin for checking the video object detection.
  """

  def _draw_witness_image(self, img, inferences, **kwargs):
    """
    Draw the inferences on the image.
    """
    for inference in inferences:
      box_tlbr = inference[self.consts.TLBR_POS]
      lbl = inference[self.consts.TYPE]
      lbl += f" | {inference.get(self.consts.COLOR_TAG)}"

      img = self._painter.draw_detection_box(
        image=img,
        top=box_tlbr[0],
        left=box_tlbr[1],
        bottom=box_tlbr[2],
        right=box_tlbr[3],
        label=lbl,
        prc=inference[self.consts.PROB_PRC],
        color=self.consts.DARK_GREEN
      )
    return img

  def process(self):
    instance_inferences = self.dataapi_image_instance_inferences()
    payload = None

    if self.cfg_debug_detections:
      self.P(f"Currently found {len(instance_inferences)} inferences.")

      if len(instance_inferences) > 0:
        self.P(f'inference keys: {instance_inferences[0].keys()}')
    # endif debug_detections

    payload_kwargs = {
      "objects": instance_inferences,
    }

    if self.cfg_demo_mode:
      np_witness = self.get_witness_image(
        draw_witness_image_kwargs={
          "inferences": instance_inferences,
        }
      )

      payload_kwargs["img"] = np_witness
    # endif demo_mode
    # The timestamp will be added to the payload in the base plugin.
    payload = self._create_payload(**payload_kwargs)
    return payload





import os

from naeural_core import Logger
from collections import OrderedDict

NAMES = {
  'buc': [
    '1.1 BUC',
    '1.2 BUC+FP',
    '1.3 BUC+FP+PER',
    '1.4 BUC+FP+PER+QUE'
    ],
  'disp': [
    '2.1 DISP',
    '2.2 DISP+FP',
    '2.3 DISP+FP+PQ',
    '2.4 DISP+FP+PQ+CNT'
    ],
  'int': [
    '3.1 INT',
    '3.2 INT+QUE'
    ],
  'hol': [
    '4.1 HOL'
    ],
  'open': [
    '5.1 OPEN',
    '5.2 OPEN+FP',
    '5.3 OPEN+FP+PER'
    ]
  }

if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(lib_name='DEVICE', config_file=cfg_file, max_lines=1000)
  
  STREAM_PATH = os.path.join('xperimental', 'streams', 'gts')
  
  all_stages = OrderedDict()
  for stream in NAMES.keys():
    path_folder = os.path.join(STREAM_PATH, stream)
    os.makedirs(path_folder, exist_ok=True)
    
    path_source = os.path.join(STREAM_PATH, stream + '.txt') 
    stream_json = log.load_json(path_source)
    
    lst_plugins = stream_json['PLUGINS']    
    
    stream_stages = {}
    for i in range(1, len(lst_plugins) + 1):
      l = lst_plugins[:i]
      crt_json = stream_json.copy()
      crt_json['PLUGINS'] = l
      save_name = NAMES[stream][i-1]
      out_path = os.path.join(path_folder, save_name + '.txt')
      log.save_json(crt_json, out_path)
      stream_stages[save_name] = crt_json
    #endfor
    out_path = os.path.join(path_folder, stream + '_stream_stages' + '.txt')
    log.save_json(stream_stages, out_path)
    all_stages.update(stream_stages)
  #endfor
  out_path = os.path.join(STREAM_PATH, 'all_stages' + '.txt')
  log.save_json(all_stages, out_path)
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  


import redis
from abc import abstractmethod
import json
from time import sleep
from copy import deepcopy

NEW_ENTITY = 'NEW'

class AbstractStateSubscriber(object):
  """
  Abstract class used to inherit state subscriber classes employed for
  the publish-subscribe design pattern.

  Works only with redis.

  Public methods:
  - subscribe(channel)
  - get_state(channel, batch_size)
  """

  def __init__(self, host=None, port=None, db=None, password=None):
    """
    Instantiates the redis subscriber object, the channels object (on
    which the message information is stored), and the default values for an empty channel

    Parameters:
    ------------
    host: str, optional
      the redis server host
      Default to None ('localhost')

    port: int, optional
      the redis server port
      Default to None (6379)

    db: int, optional
      the redis server db
      Default to None (0)

    password: str, optional
      The password of the redis server
      Default to None - if the server does not have any password
    """
    if host is None:
      host = 'localhost'

    if port is None:
      port = 6379

    if db is None:
      db = 0

    self.default_value_channel = None
    self._set_default_value_channel()
    self.channels = dict()

    self.redis_ssp = redis.Redis(
      host=host, port=port, db=db,
      password=password,
      health_check_interval=10
    )
    self._subscriber = self.redis_ssp.pubsub()
    return

  @abstractmethod
  def _refresh_messages(self, batch_size=None):
    """
    Abstract method used for iterating through the redis queue and storing the
    published messages

    Parameters:
    -----------
    batch_size: int, optional
      number of maximum messages to be processed one time. Used to handle
      the situation in which the messages are sent faster than the method can handle.
      Default to None


    Implementation info:
    - this method iterates through the message queue, decodes the jsons, and loads them into the
    `self.channels` object
    - If the message queue is too long it only processes the first 'batch_size' messages. This
    is done in order to handle the situation in which the messages are sent faster than the
    method can handle.
    """
    pass

  @abstractmethod
  def _set_default_value_channel(self):
    """
    Abstract method to be implemented in child classes.
    Sets `self.default_value_channel` (the initial value for `self.channels[channel]`)
    to a custom value
    """
    pass

  def subscribe(self, channel):
    """
    Method used for subscribing to a redis channel

    Parameters:
    -----------
    channel: str, mandatory
      the channel to subscribe to
    """
    if channel not in self.channels:
      self.channels[channel] = deepcopy(self.default_value_channel)

    self._subscriber.subscribe(channel, socket_keepalive=True)

  def _clear_state(self, channel):
    """
    Method used to reset the data stored for a given channel

    Parameters:
    ----------
    channel: str, mandatory
      the channel to clear the state
    """
    self.channels[channel] = deepcopy(self.default_value_channel)

  def get_state(self, channel=None, batch_size=None):
    """
    Method used to get the queued messages from a channel. It calls `refresh_messages`
    method, clears the queue and returns the messages

    Parameters:
    -----------
    channel: str, optional
      the channel from which to get the messages from
      Default is None (the first channel found in `channels` object)

    batch_size: int, optional
      number of maximum messages to be processed one time.
      Default is None
    """
    if channel is None:
      if len(self.channels) > 0:
        channel = list(self.channels.keys())[0]

    self._refresh_messages(batch_size=batch_size)
    state = self.channels[channel]
    self._clear_state(channel)
    return state



class SingleMessageStateSubscriber(AbstractStateSubscriber):
  """
  Class defined to implement the logic for the case when each channel receives a single message.

  Public methods:
    - subscribe(channel)
    - get_state(channel, batch_size)
  """

  def _set_default_value_channel(self):
    """
    Sets `self.default_value_channel` to None - each channel receives a single message.

    See `AbstractStateSubscriber._set_default_value_channel` for more docstring
    """
    self.default_value_channel = None

  def _refresh_messages(self, batch_size=None):
    """
    Method used for iterating through the redis queue and storing the published messages.
    See `AbstractStateSubscriber._refresh_messages` for more docstring

    Parameters:
    -----------
    batch_size: int, optional
      number of maximum messages to be processed one time. Used to handle
      the situation in which the messages are sent faster than the method can handle.
      Default to None (1000)
    """

    if batch_size is None:
      batch_size = 1000

    if len(self.channels) == 0:
      return

    for _ in range(batch_size):
      msg = self._subscriber.get_message()

      if msg is None:
        break

      try:
        channel = msg['channel'].decode('utf-8')
        data = msg['data'].decode('utf-8')
      except:
        continue

      self.channels[channel] = data
      sleep(0.01)
    # end - for _ in range(batch_size):

    return


class MultipleMessagesStateSubscriber(AbstractStateSubscriber):
  """
  Class defined to implement the logic for the case when each channel receives multiple messages,
  indexed in a dictionary

  Public methods:
    - subscribe(channel)
    - get_state(channel, batch_size)
  """

  def __init__(self, channel, **kwargs):
    super().__init__(**kwargs)
    self.subscribe(channel)
    self.channels[channel] = deepcopy(self.default_value_channel)
    return

  def _set_default_value_channel(self):
    """
    Sets `self.default_value_channel` to `{}` - each channel receives multiple messages,
    indexed in a dictionary
    """
    self.default_value_channel = {}

  def _refresh_messages(self, batch_size=None):
    """
    Method used for iterating through the redis queue and storing the published messages.
    See `AbstractStateSubscriber._refresh_messages` for more docstring

    Parameters:
    -----------
    batch_size: int, optional
      number of maximum messages to be processed one time. Used to handle
      the situation in which the messages are sent faster than the method can handle.
      Default to None (1000)
    """
    if batch_size is None:
      batch_size = 1000

    if len(self.channels) == 0:
      return

    for _ in range(batch_size):
      msg = self._subscriber.get_message()

      if msg is None:
        break

      try:
        channel = msg['channel'].decode('utf-8')
        data = msg['data'].decode('utf-8')
      except:
        continue

      data = json.loads(
        data,
        object_hook=lambda d: {int(k) if k.isnumeric() else k: v for k, v in d.items()}
      )

      object_id = list(data.keys())[0]
      obj_state = data[object_id]


      if object_id != NEW_ENTITY:
        if object_id not in self.channels[channel]:
          self.channels[channel][object_id] = {}

        for key_state, value_state in obj_state.items():
          self.channels[channel][object_id][key_state] = value_state

      else:
        if object_id not in self.channels[channel]:
          self.channels[channel][object_id] = []

        self.channels[channel][object_id].append(obj_state)
      #endif

      sleep(0.01)
    # end - for _ in range(batch_size):

    return



import cv2
import time
from numba import jit
import matplotlib.pyplot as plt

@jit(nopython=False)
def resize_img(img):
  img = cv2.resize(img, (1920, 1080))
  return img

if __name__ == '__main__':
  img = cv2.imread('xperimental/covid2.jpg')
  plt.imshow(img)
  start = time.time()
  img = resize_img(img)
  print(time.time() - start)



import traceback
import os

from time import time


class _OrchestratorUtils(object):
  def __init__(self):
    super(_OrchestratorUtils, self).__init__()
    return
  
  
  def _save_object_tree(self, fn='obj_mem_tree.txt', save_top_only=True, top=100):   
    self.P("  Logging startup object tree in '{}' ...".format(fn), color='r')
    t3 = time()
    try:
      mem2, tree2, top2 = self.log.get_obj_size(
        obj=self, 
        top_consumers=top,
        return_tree=True,
      )
    except:
      mem2, tree2, top2  = 0, 0, []
      self.P(" * * * * * * * * * * * * ERROR: get_obj_size FAILED !  * * * * * * * * * * * *\n{}".format(
        traceback.format_exc()
      ))
    #endtry    
    t4 = time()
    mem, tree, top = mem2, tree2, top2 
    self.P("  Done v2 logging object tree: {:,.1f} MB.".format(mem/1024**2), color='r')
    elapsed = t4 - t3
    if tree is not None:
      folder = self.log.get_output_subfolder('memory_logs', force_create=True)
      self.log.maybe_cleanup_timestamped_files(folder=folder, keep=8)
      if not save_top_only:
        _fn = self.log.now_str(short=True)[:12] + '_v2_' + fn
        with open(os.path.join(folder, _fn), 'wt') as fh:
          for line in tree:
            fh.write('{}\n'.format(line))
          
      _fn = self.log.now_str(short=True)[:12] + '_v2_top_' + fn
      with open(os.path.join(folder, _fn), 'wt') as fh:
        for line in top:
          fh.write('{}\n'.format(line))
          
      self.P("  Heavy duty op finished: logging startup object tree in {:.3f}s.".format(elapsed), color='r')      
      display = "\n".join(["(L{}) {}: {:.1f}".format(x['LEVEL'], x['NAME'], x['SIZE'] / 1024**2) for x in top])
      self.P("  Showing top {} memory consumers in MB:\n{}\n".format(len(top), display))
    else:
      self.P("  ERROR: Logging failed", color='r')
    #endif 
    return  


import numpy as np
import abc
from naeural_core.local_libraries.nn.th.training.callbacks.classification import ClassificationTrainingCallbacks

class SoftmaxClassificationTrainingCallbacks(ClassificationTrainingCallbacks, abc.ABC):
  def __init__(self, **kwargs):
    super(SoftmaxClassificationTrainingCallbacks, self).__init__(**kwargs)
    return

  def _get_y(self, lst_y):
    y = np.hstack(lst_y)
    return y

  def _get_y_hat(self, lst_y_hat):
    y_hat = np.vstack(lst_y_hat)
    y_hat = y_hat.argmax(axis=1)
    y_hat = y_hat.reshape(-1)
    return y_hat



from naeural_core.comm import AMQPWrapper
from naeural_core.core_logging import SBLogger
from collections import deque
from time import sleep

if __name__ == '__main__':

  log = SBLogger()
  url_config_app = 'https://www.dropbox.com/s/h4x9dd1pim92s94/config_app_pika_demo.txt?dl=1'
  # url_config_app = 'https://www.dropbox.com/s/ex1zronpiz8bobk/config_app_pika_staging.txt?dl=1'
  saved_files, _ = log.maybe_download(
    url=url_config_app,
    fn='tmp/config_app_pika.txt',
    force_download=True,
    target='output'
  )

  config_app = log.load_json(saved_files[0])
  config_communication = config_app['COMMUNICATION']

  log.P("Server params:", color='b')
  for k, v in config_communication['PARAMS'].items():
    log.P(" * {}: {}".format(k, v), color='b')

  config_communication['PARAMS']['EE_ID'] = 'icsulescu'

  recv_buffer_1 = deque(maxlen=1000)
  recv_buffer_2 = deque(maxlen=1000)
  sender = AMQPWrapper(log=log, config=config_communication['PARAMS'],
                       recv_buff=None, send_channel_name='PAYLOADS_CHANNEL', DEBUG=True)
  receiver1 = AMQPWrapper(
    log=log, config=config_communication['PARAMS'], recv_buff=recv_buffer_1, recv_channel_name='PAYLOADS_CHANNEL')
  receiver2 = AMQPWrapper(
    log=log, config=config_communication['PARAMS'], recv_buff=recv_buffer_1, recv_channel_name='PAYLOADS_CHANNEL')

  _dct_conn_sender = sender.server_connect()
  log.P("Send  conn: {}".format(_dct_conn_sender))
  _dct_conn_receiver1 = receiver1.server_connect()
  log.P("Recv1 conn: {}".format(_dct_conn_receiver1))
  _dct_conn_receiver2 = receiver2.server_connect()
  log.P("Recv2 conn: {}".format(_dct_conn_receiver2))

  _dct_publish_sender = sender.establish_one_way_connection('send')
  log.P("Send  pub : {}".format(_dct_publish_sender))
  _dct_subscribe_receiver1 = receiver1.establish_one_way_connection('recv')
  log.P("Recv1 subs: {}".format(_dct_subscribe_receiver1))
  _dct_subscribe_receiver2 = receiver2.establish_one_way_connection('recv')
  log.P("Recv2 subs: {}".format(_dct_subscribe_receiver2))

  def read_all(receiver, receiver_name):
    log.P("Flushing the queue for receiver {} ...".format(receiver_name))
    prev_len = len(receiver._recv_buff)
    while True:
      receiver.receive()
      crt_len = len(receiver._recv_buff)
      if crt_len == prev_len:
        break

      prev_len = crt_len
      sleep(0.1)
    # endwhile

    log.P("  Flush done, there were {} messages in the queue:".format(crt_len))
    while len(receiver._recv_buff) > 0:
      x = receiver._recv_buff.popleft()
      log.P("  {}".format(x))
  # enddef

  read_all(receiver1, '1')
  read_all(receiver2, '2')

  for i in range(30):
    iteration = i + 1
    log.P("\n====Iteration {:03d}====".format(iteration))
    sender.send(log.safe_dumps_json({'msg_id': i + 1}, ensure_ascii=False))
    read_all(receiver1, '1')
    read_all(receiver2, '2')
  # endfor



import numpy as np
from collections import OrderedDict

class _FitDebugTFKerasMixin(object):
  """
  Mixin for tf keras fit debug functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_TimersMixin`:
    - self.start_timer
    - self.end_timer
  """

  def __init__(self):
    super(_FitDebugTFKerasMixin, self).__init__()

    try:
      from ratio1.logging.logger_mixins.timers_mixin import _TimersMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _FitDebugTFKerasMixin without having _TimersMixin")

    return

  def fit_debug(self, model, x_train, y_train,
                first_iters=1, max_grad=10, max_weight=10,
                validation_data=None,
                batch_size=32, epochs=0, mode='keras'):
    from tensorflow.python.eager import context
    import tensorflow as tf

    K = tf.keras.backend
    mode = mode.upper()
    optimizer = tf.keras.optimizers.get(model.optimizer)
    loss = model.loss
    output_names = [x.name.split('/')[0] for x in model.outputs]
    # Prepare loss functions.
    if isinstance(loss, dict):
      for name in loss:
        if name not in output_names:
          raise ValueError(
            'Unknown entry in loss '
            'dictionary: "' + name + '". '
                                     'Only expected the following keys: ' + str(self.output_names))
      loss_functions = []
      for name in output_names:
        if name not in loss:
          self.P(
            'Output "' + name + '" missing from loss dictionary. '
                                'We assume this was done on purpose, '
                                'and we will not be expecting '
                                'any data to be passed to "' + name + '" during training.')
        loss_functions.append(tf.keras.losses.get(loss.get(name)))
    elif isinstance(loss, list):
      if len(loss) != len(model.outputs):
        raise ValueError('When passing a list as loss, '
                         'it should have one entry per model outputs. '
                         'The model has ' + str(len(model.outputs)) +
                         ' outputs, but you passed loss=' + str(loss))
      loss_functions = [tf.keras.losses.get(l) for l in loss]
    else:
      loss_function = tf.keras.losses.get(loss)
      loss_functions = [loss_function for _ in range(len(output_names))]

    if mode == 'KERAS':
      x_tensors = model.inputs
      y_tensors = []
      for output in model.outputs:
        y_tensors.append(tf.keras.layers.Input(batch_shape=output.shape))
      yh_tensors = model.outputs

      tf_loss = 0
      for i, loss_func in enumerate(loss_functions):
        tf_temp_loss = loss_func(y_tensors[i], yh_tensors[i])
        tf_loss += K.mean(tf_temp_loss)

      tf_grads = K.gradients(tf_loss, model.trainable_variables)
      get_updates = optimizer.get_updates(loss=tf_loss,
                                          params=model.trainable_variables)
      train_func = K.function(inputs=x_tensors + y_tensors,
                              outputs=[tf_loss] + tf_grads,
                              updates=get_updates)
      loss_func = K.function(inputs=x_tensors + y_tensors,
                             outputs=[tf_loss])
    elif mode == 'TF':
      raise ValueError("Not implemented yet")
    elif mode == 'EAGER':
      if not context.executing_eagerly():
        raise ValueError("Cannot debug in EAGER mode without eager execution!")
      raise ValueError("Not implemented yet")

    if isinstance(x_train, list):
      train_size = x_train[0].shape[0]
    elif isinstance(x_train, np.ndarray):
      train_size = x_train.shape[0]
      x_train = [x_train]

    if not isinstance(y_train, list):
      y_train = [y_train]

    iters = range(train_size // batch_size + 1) if first_iters in [None, 0] else range(first_iters)
    s_epochs = "{:>" + str(len(str(epochs)) + 1) + "}"
    prints = 20
    print_every = max(len(iters) // prints, 1)
    self.P("Debug training for {} epochs and {} iters".format(
      epochs, iters))
    epochs = max(epochs, 1)
    for epoch in range(epochs):
      _s = s_epochs.format(epoch + 1)
      self.P("Training epoch {}...".format(_s))
      self.start_timer("DEBUG_FIT_LOGGER_EPOCH")
      for b in iters:
        start = b * batch_size
        end = start + batch_size
        if end > train_size:
          end = train_size
        if not (end > start):
          break
        x_batch = [x[start:end] for x in x_train]
        y_batch = [y[start:end] for y in y_train]

        res = train_func(x_batch + y_batch)
        loss_after = loss_func(x_batch + y_batch)[0]
        loss = res[0]
        grads = res[1:]
        if (b % print_every) == 0:
          print("\rTraining batches {:.1f}% loss: {:.5f} loss (post GD): {:.5f}".format(
            b / len(iters) * 100, loss, loss_after
          ), end="", flush=True)
      # epoch end
      print("")
      self.debug_weights_grads(grads, model=model, max_grad=max_grad, max_weight=max_weight)
      elapsed = self.end_timer("DEBUG_FIT_LOGGER_EPOCH")
      self.P("Done epoch {} in {:.1f}s".format(_s, elapsed))
      if validation_data:
        results = model.evaluate(validation_data[0], validation_data[1], verbose=0)
        self.P("Epoch {} validation results:".format(_s))
        for i, metric in enumerate(model.metrics_names):
          self.P(" Val {:<14} {:.3f}".format(metric[:13] + ":", results[i]))
      else:
        results = model.evaluate(x_batch, y_batch, verbose=0)
        if not isinstance(results, list):
          results = [results]
        self.P("Epoch {} training results (no val data) on last batch:".format(_s))
        for i, metric in enumerate(model.metrics_names):
          self.P(" Train {:<14} {:.3f}".format(metric[:13] + ":", results[i]))

    return

  def debug_weights_grads(self, grads, model, max_grad=10, max_weight=10):
    import tensorflow.keras.backend as K
    layers_info = OrderedDict()
    lr = K.eval(model.optimizer.lr)
    var_names = []
    for i, _var in enumerate(model.trainable_variables):
      _name = _var.name.split(":")[0]
      l_name, w_name = _name.split("/")
      var_name = l_name[:15] + "/" + w_name
      var_name = var_name[:19]
      var_names.append(var_name)
    weights = model.get_weights()
    self.P(" Checking weights for model with lr={:.1e}".format(lr))
    for i, var_name in enumerate(var_names):
      _var = model.trainable_variables[i]
      var_shape = K.int_shape(_var)
      w = weights[i]
      _min = w.min()
      _max = w.max()
      _med = np.median(w)
      _avg = w.mean()
      _norm = np.linalg.norm(w)
      has_nans = np.isnan(w).any()
      aprox_zero = (w <= K.epsilon()) & (w >= -K.epsilon())
      prc_aprox_zero = aprox_zero.sum() / aprox_zero.size
      nr_big = (w >= max_weight) | (w <= -max_weight)
      prc_big = nr_big.sum() / nr_big.size
      has_issues = False
      if has_nans:
        s_issue = "  NaN weights !!!"
        has_issues = True

      if prc_aprox_zero >= 0.25:
        s_issue = "    {:.1f}% zero-weights !".format(prc_aprox_zero * 100)
        has_issues = True

      if prc_big >= 0.25:
        s_issue = "    {:.1f}% big weights !".format(prc_big * 100)
        has_issues = True

      s1 = str(var_shape)
      check = 'OK' if not has_issues else 'NOT OK'
      self.P("  Weights for  {:<19} {:<15}: {}".format(var_name, s1, check))
      if has_issues:
        self.P(s_issue)
        self.P("    Norm:    {:>8.1e}".format(_norm))
        self.P("    Range:   {:>8.1e} - {:>8.1e}".format(_min, _max))
        self.P("    Avg/Med: {:>8.1e} / {:>8.1e}".format(_avg, _med))

    self.P(" Debugging gradients")
    for i, var_name in enumerate(var_names):
      _var = model.trainable_variables[i]
      var_shape = K.int_shape(_var)
      grad = grads[i]
      has_nans = np.isnan(grad).any()
      aprox_zero = (grad <= K.epsilon()) & (grad >= -K.epsilon())
      prc_aprox_zero = aprox_zero.sum() / aprox_zero.size
      nr_big = (grad >= max_grad) | (grad <= -max_grad)
      prc_big = nr_big.sum() / nr_big.size
      has_issues = False

      if has_nans:
        s_issue = "  NaN gradient !!!"
        has_issues = True

      if prc_aprox_zero >= 0.25:
        s_issue = "    {:.1f}% zero-grads !".format(prc_aprox_zero * 100)
        has_issues = True

      if prc_big >= 0.25:
        s_issue = "    {:.1f}% big grads !".format(prc_big * 100)
        has_issues = True

      _min = grad.min()
      _max = grad.max()
      _med = np.median(grad)
      _avg = grad.mean()
      _norm = np.linalg.norm(grad)
      layer = {"shape": var_shape, "grad_min": _min, "grad_norm": _norm}
      check = 'OK' if not has_issues else 'NOT OK'
      s1 = str(var_shape)
      self.P("  Grads for  {:<19} {:<15}: {}".format(var_name, s1, check))
      if has_issues:
        self.P(s_issue)
        self.P("    Norm:    {:>8.1e}".format(_norm))
        self.P("    Range:   {:>8.1e} - {:>8.1e}".format(_min, _max))
        self.P("    Avg/Med: {:>8.1e} / {:>8.1e}".format(_avg, _med))
      layers_info[var_name] = layer
    return layers_info



"""
Example pipeline configuration:

{
    "NAME": "on-demand-stream",
    "TYPE": "OnDemandTextInput",

    "PLUGINS": [
        {
            "SIGNATURE": "code_assist_01"
            "INSTANCES": [
                {
                    "INSTANCE_ID": "default"
                }
            ],
        }
    ],
    "STREAM_CONFIG_METADATA" : {
      "SERVER_SIDE_PARAMS_HERE" : "1234"
    }
}

then:

{ 
  "ACTION" : "PIPELINE_COMMAND",  
  "PAYLOAD" : {
    "NAME": "on-demand-stream",
    "PIPELINE_COMMAND" : {
      "STRUCT_DATA" : ["hello world"]
    }
  }
}
"""

from naeural_core.data.base import DataCaptureThread

__VER__ = '0.1.0'

_CONFIG = {
  **DataCaptureThread.CONFIG,
  
  'CAP_RESOLUTION'    : 10, # default run at 10Hz not that it really matters :)
  
  'LIVE_FEED'         : False, # it should run buffered
  
  'VALIDATION_RULES'  : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}

class OnDemandTextInputDataCapture(
  DataCaptureThread,

  ):

  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(OnDemandTextInputDataCapture, self).__init__(**kwargs)
    return

  
  def startup(self):
    # NON-threaded code in startup
    super().startup()
    self._metadata.update(
      pipeline_version=__VER__,
    )
    return 
  
  
  def _init(self):
    self._iter = 0
    # not much here
    self._maybe_reconnect()    
    return
  
  @property
  def is_idle_alert(self):
    # we are never idle :)
    return False
  
  
  def _maybe_reconnect(self):
    # not much here
    if self.has_connection:
      return    
    self.has_connection = True
    return  
  
  
  def _on_pipeline_command(self, cmd_data, payload_context=None, **kwargs):
    _obs = cmd_data.get('STRUCT_DATA', [])
    _train = None
    if not isinstance(_obs, list):
      _obs = [_obs]
      
    inputs = [
        self._new_input(img=None, struct_data=x, metadata=self._metadata.__dict__.copy(), init_data=_train)
        for x in _obs
    ]
    self._add_inputs(inputs)
    return  
    
    
  def _run_data_aquisition_step(self):
    self._iter += 1
    # not much here either as stuff happens in _on_pipeline_command
    return 
    
    
  def _release(self):    
    return  
  
  


import numpy as np
import torch as th
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

if __name__ == "__main__":
    img1 = np.zeros((600,400,3), dtype=np.uint8)
    img2 = np.zeros((1080,1920,3), dtype=np.uint8)
    
    th_sub = th.tensor([104, 117, 123]).unsqueeze(-1).unsqueeze(-1).to(th.device("cuda"))
    
    imgs = [img1, img2]
    th_x, dims = th_resize_with_pad(
        img=imgs, 
        h=720,
        w=1280,
        device=th.device("cuda"), 
        div_val=1,
        sub_val=th_sub, 
    )
    
    print("dev")
    


from time import sleep
import threading
import ctypes

NULL = 0
dct_shmem = {
  }
        

def ctype_async_raise(target_tid, exception):
    ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(target_tid), ctypes.py_object(exception))
    # ref: http://docs.python.org/c-api/init.html#PyThreadState_SetAsyncExc
    if ret == 0:
      raise ValueError("Invalid thread ID")
    elif ret > 1:
      # Huh? Why would we notify more than one threads?
      # Because we punch a hole into C level interpreter.
      # So it is better to clean up the mess.
      ctypes.pythonapi.PyThreadState_SetAsyncExc(target_tid, NULL)
      raise SystemError("PyThreadState_SetAsyncExc failed")
    print("Successfully set asynchronized exception for", target_tid)

def main_stopper():
  sleep(dct_shmem['wait'])
  ctype_async_raise(dct_shmem['pid'], ValueError)
  return


if __name__ == '__main__':
  dct_shmem['pid'] = threading.get_ident()
  dct_shmem['wait'] = 10
  
  thr = threading.Thread(target=main_stopper, daemon=True)
  thr.start()
  try:
    while True:
      sleep(0.1)
  except Exception as e:
    print('Exited from forever loop: {}'.format(e))
  
  print('Done.')
  
  
  


import abc
import json

from naeural_core import DecentrAIObject
from naeural_core import Logger
from naeural_core import constants as ct


_CONFIG = {
  'VALIDATION_RULES' : {

  }
}

class BaseConfigRetrievingPlugin(DecentrAIObject):

  CONFIG = _CONFIG
  __metaclass__ = abc.ABCMeta

  def __init__(self, log : Logger, config, **kwargs):
    self.config = config
    super(BaseConfigRetrievingPlugin, self).__init__(log=log, **kwargs)
    return
  
  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color,str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.MAIN
    super().P(s, prefix=False, color=color, **kwargs)
    return    

  def startup(self):
    super().startup()
    self.config_data = self.config
    return

  def retrieve(self, app_config_endpoint=None, streams_configs_endpoint=None):
    config_app, lst_config_streams = None, None
    dct_config_streams = None
    if app_config_endpoint is not None:
      self.log.start_timer('get_app_configuration')
      try:
        self.P("  Running `_get_app_configuration`...")
        config_app = self._get_app_configuration(app_config_endpoint)
        self.P("App config - Successfully retrieved from endpoint '{}'".format(app_config_endpoint), color='g')
      except Exception as e:
        msg = "App config - ERROR! Could not retrieve from endpoint '{}'".format(app_config_endpoint)
        self.P(msg, color='r')
        self._create_notification(notif=ct.STATUS_TYPE.STATUS_EXCEPTION, msg=msg, autocomplete_info=True)
      #end try-except
      self.log.stop_timer('get_app_configuration', skip_first_timing=False)
    #endif

    if streams_configs_endpoint is not None:
      try:
        self.log.start_timer('get_streams_configurations')
        lst_config_streams = self._get_streams_configurations(streams_configs_endpoint)
        self.log.stop_timer('get_streams_configurations', skip_first_timing=False)
        self.P("Streams configs - Successfully retrieved from endpoint '{}'".format(streams_configs_endpoint), color='g')
      except Exception as e:
        msg = "Streams configs - ERROR! Could not retrieve from endpoint '{}'".format(streams_configs_endpoint)
        self.P(msg, color='r')
        self._create_notification(notif=ct.STATUS_TYPE.STATUS_EXCEPTION, msg=msg, autocomplete_info=True)
      #end try-except

      if lst_config_streams:
        good_streams, good_streams_fnames = self._keep_good_streams(lst_config_streams)
        if len(good_streams_fnames) > 0:
          self.P("  New streams: {}".format("; ".join(good_streams_fnames)), color='g')
        lst_config_streams = good_streams
        dct_config_streams = {s['NAME'] : s for s in lst_config_streams}
      #endif
    #endif

    return config_app, dct_config_streams

  def connect(self, **kwargs):
    try:
      self._connect(**kwargs)
      self.P("Config retriever {} connection succeeded".format(self.__class__.__name__), color='g')
    except Exception as e:
      msg = "Config retriever '{}' connection failed".format(self.__class__.__name__)
      self.P(msg, color='r')
      self._create_notification(notif=ct.STATUS_TYPE.STATUS_EXCEPTION, msg=msg, autocomplete_info=True)
    #end try-except

    return



  def _keep_good_streams(self, lst_config_streams):
    good_streams, bad_streams = [], []
    fnames = []
    for i,config_s in enumerate(lst_config_streams):
      name = config_s.get('NAME', None)
      if name is None:
        bad_streams.append(config_s)
      else:
        good_streams.append(config_s)
        fnames.append(name + '.txt')
    #endfor

    if len(bad_streams) > 0:
      msg = "ERROR! {}/{} retrieved streams are bad configured - have no 'NAME'. Therefore they will not be started".format(
        len(bad_streams), len(lst_config_streams)
      )
      info = json.dumps(bad_streams)
      self.P(msg + "\n" + info, color='r')
      self._create_notification(notif=ct.STATUS_TYPE.STATUS_EXCEPTION, msg=msg, info=info)
    #endif

    return good_streams, fnames



  ### START - MUST BE OVERWRITTEN
  def _connect(self, **kwargs):
    """
    This method should be implemented when the plugin has to create a connection to an API / DB etc.

    Parameters:
    ----------
    kwargs:
      Connection arguments

    Returns:
    --------
    None
    """
    raise NotImplementedError()
  ### END - MUST BE OVERWRITTEN

  ### START - MUST BE IMPLEMENTED
  @abc.abstractmethod
  def _get_app_configuration(self, endpoint):
    """
    Parameters:
    -----------
    endpoint
      The endpoint from where the app configuration is retrieved.
      The type of this parameter depends on the plugin

    Returns:
    --------
    config_app : dict
      The application configuration that will be saved on the local device data storage (_local_cache/_data).
      The reason why we return dict is that the public function `retrieve` handles saving.
    """
    raise NotImplementedError()


  @abc.abstractmethod
  def _get_streams_configurations(self, endpoint):
    """
    Parameters:
    -----------
    endpoint
      The endpoint from where the streams configurations are retrieved.
      The type of this parameter depends on the plugin

    Returns:
    --------
    lst_config_streams : List[dict]
      A list containing the streams configurations that will be saved on the local device data storage  (_local_cache/_data).
      The reason why we return List[string] is that the public function `retrieve` handles saving.
    """
    raise NotImplementedError()


  ### END - MUST BE IMPLEMENTED



import asyncio
from functools import wraps

class _MultithreadingMixin(object):
  """
  Mixin for multithreading functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_MultithreadingMixin, self).__init__()
    return

  @staticmethod
  def background(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
      loop = asyncio.get_event_loop()
      if callable(f):
        return loop.run_in_executor(None, f, *args, **kwargs)
      else:
        raise TypeError('Task must be a callable')
      #endif

    return wrapped

  @staticmethod
  def parallelise(function, data, batch=100, verbose=True):
    """
    Parameters
    ----------
    function : callback
      Function you need to run in parallel, on cpu, in order to speed up calculations. Ex: f2_score, etc
    data : List
      List containing tuples or values you need to pass to function
    batch : Integer, optional
      Data allocation of per job (Remember! A low batch increases IPC and synchronization overhead). The default is 1000.
    verbose: Boolean, optional
      Printing current status. The default is True.
    Returns
    -------
    results : List
      List of function result.
    """
    from multiprocessing import cpu_count, Pool
    from tqdm import tqdm
    if verbose:
      print('Parallelise function {} with data of length {} and batch {}'.format(function.__name__, len(data), batch))
    pool = Pool(processes=cpu_count())
    results = list(tqdm(pool.imap(function, data, batch), total=len(data), disable=not verbose))
    pool.close()
    pool.join()
    if verbose:
      print('Done parallelise function {}'.format(function.__name__))
    return results



import numpy as np

from naeural_core.utils.basic_anomaly_model import BasicAnomalyModel

ANOMALY_THRESHOLD = 0.02

if __name__ == '__main__':
    model = BasicAnomalyModel()
    x_train = np.random.randint(low=4, high=7, size=10000)
    x_train = x_train.reshape(-1, 1)
    model.fit(x_train=x_train, prc=ANOMALY_THRESHOLD)

    # model.predict(x_test=)
